/* ======================== program MinMax ======================
   Program to find minima and maxima in a data file
   (c) Petr Mikulik, Brno, 1995, 1998, 1999
   mikulik@physics.muni.cz, http://www.sci.muni.cz/~mikulik/

   This file and all included files can be used (distributed, modified)
   freely as long as the copyright notice is left intact.
   It is distributed without any warranty.

   Original version: May 1995
   Bugfix release: April 1998 (sometimes did not work well for descending sequences)
   Support for input pipes and stdin: May 1999
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <fstream.h>
#include <iomanip.h>
#include "ansicpp.h"
#include "pmparser.h"
#include "lfstream.h"
#include "theline.h"

#define TheDate "May 1999"

int SkipLines = 0;



// ############ Initialization and run fitting ##############

void Make (char *InputFile)
{
// read file and calculate min, max
int i;
double XX, YY, ZZ;
double MinX=1e300,Min2X=1e300,
       MinY=1e300,Min2Y=1e300,
       MinZ=1e300,Min2Z=1e300;
double MaxX=-1e300,Max2X=-1e300,
       MaxY=-1e300,Max2Y=-1e300,
       MaxZ=-1e300,Max2Z=-1e300;

cout <<"Reading file " <<InputFile <<'\n' COUT_FLUSH;
lfstream Inp;
Inp.open( InputFile );

for (i=SkipLines; i-->0; Inp.xligno());
ThisEveryNth = EveryNth = 1;

while (!Inp.eof_skipspaces()) {
  ReadTheLine(XX,YY,ZZ,Inp,1);
  if (Inp.fail()) break;

  if (XX<MinX) { Min2X=MinX; MinX=XX; } else
    if (XX<Min2X) { if (XX>MinX) Min2X=XX; }
  if (XX>MaxX) { Max2X=MaxX; MaxX=XX; } else
    if (XX>Max2X && XX<MaxX) Max2X=XX;

  if (YY<MinY) { Min2Y=MinY; MinY=YY; } else
    if (YY<Min2Y) { if (YY>MinY) Min2Y=YY; }
  if (YY>MaxY) { Max2Y=MaxY; MaxY=YY; } else
    if (YY>Max2Y && YY<MaxY) Max2Y=YY;

  if (ColumnZ) {
    if (ZZ<MinZ) { Min2Z=MinZ; MinZ=ZZ; } else
      if (ZZ<Min2Z) { if (ZZ>MinZ) Min2Z=ZZ; }
    if (ZZ>MaxZ) { Max2Z=MaxZ; MaxZ=ZZ; } else
      if (ZZ>Max2Z && ZZ<MaxZ) Max2Z=ZZ;
    }

  } // while not eof

Inp.close();  // end of reading

if (Max2X<Min2X) Max2X=Min2X;
if (MaxX<Max2X) MaxX=Max2X;
if (Max2Y<Min2Y) Max2Y=Min2Y;
if (MaxY<Max2Y) MaxY=Max2Y;
if (Max2Z<Min2Z) Max2Z=Min2Z;
if (MaxZ<Max2Z) MaxZ=Max2Z;

cout <<"** Results: **\n";
if (ColumnX>0) cout<<"Column "<<ColumnX; else cout<<FormulaX;
cout<<": min and max = \t"<<MinX<<"\t"<<Min2X<<"\t ... \t"<<Max2X<<"\t"<<MaxX<<"\n";
if (ColumnY>0) cout<<"Column "<<ColumnY; else cout<<FormulaY;
cout<<": min and max = \t"<<MinY<<"\t"<<Min2Y<<"\t ... \t"<<Max2Y<<"\t"<<MaxY<<"\n";
if (ColumnZ) {
if (ColumnZ>0) cout<<"Column "<<ColumnZ; else cout<<FormulaZ;
cout <<": min and max = \t"<<MinZ<<"\t"<<Min2Z<<"\t ... \t"<<Max2Z<<"\t"<<MaxZ<<"\n";
}
cout<<'\n' COUT_FLUSH;

}


// #################### begin of program ####################

int main( int argc, char *argv[] )
{
cout << "********* Program MinMax (version "<<TheDate<<") - c "<<Petr<<" **********\n";

if (argc<4 || argc>5) {
  cout << "Usage:  minmax input_file expr1 expr2 [expr3]\n";
  #ifdef __VMS
  cout << "On VAX/VMS systems:  MinMax:==\"$disk:[dir]MinMax.exe\"\n";
  #endif
  #if defined(__MSDOS__) && !(defined(__TURBOC__) && defined(__STDC__))
  #endif
  cout << "Examples:\n";
  cout <<" minmax data.dat x 2*y log(z+1e-3)\n";
  cout <<" minmax exp.dat '$1' '2*$2' 'log($8+exp($5))'\n";
  cout <<" (apostrophes are required on Unix systems).\n";
  cout <<"Program to find minima and maxima and data file. This program is public-domain\n"
       <<"and it is distributed together with the source code in C++.\n";
//       <<"(see ftp.muni.cz:pub/muni.cz/physics or ftp server on (www or gopher).muni.cz)\n";
  exit(1); }

ReadColumns123( argv[2],argv[3],argv[4],argc<5 ? 2:3 ); 
Make(argv[1]);
//cout<<"Program MinMax (version "<<TheDate<<") - done.\n" COUT_FLUSH;
return 0;
}

// eof minmax.cpp
