/***********************  Program  pm3d.cpp  ***********************

                *** BEGIN OF NOTES. PLEASE READ: ***

  Draw 3d data file as the gray/colour scale map. Postscript output.
  Works under OS/2, Unix, VMS, DOS; and maybe others too

  (c) Petr Mikulik - Brno - 1995--2003
  Address: mikulik@physics.muni.cz
	   http://www.sci.muni.cz/~mikulik/

	   Institute of Condensed Matter Physics
	     // Previously:
	        Laboratory of Thin Films and Nanostructures
		and
	        Department of Solid State Physics
	   Faculty of Science, Masaryk University
	   Kotlarska 2
	   611 37 Brno
	   Czech Republic

  This is version 2.9.1, October 2003
  Last change: 28. 10. 2003
  Previous changes:
	4. 7. 1998, 5.4.1998, 31.10.1997, 11. 11. 1998, 16. 1. 1999, 
	31. 3. 1999, 25. 4. 1999, 31. 10. 1999, 3. 2. 2002


*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA



  Even this program is free, it is copyrighted by me. You are not allowed
  to distributed a modified version of this program, or any other program 
  with the same name. 

  The complete distribution of this program (source code in C++, Makefiles
    compiled exe file for OS/2 and DOS, and Demo) can be downloaded from my 
    web page: http://www.sci.muni.cz/~mikulik/

  If you use DOS, you cannot use pipe to ghostscript. However, you can
    profit from the program PMGRAF.EXE (see below), where my 3d algorithm
    has been implemented for the first time. It works great also in full
    screen DOS session under OS/2, and it supports mouse zoom (and other
    features). However, there are only 16 colours available.

  Please, let me know about your opinion or recommendations for this
    program. I can put you on my mailing list if you would like to
    know about new versions. Or look to my WWW page if you want to know
    whether any new version has been released.

 Note: a nice postscript previewer for DOS (it is a batch file that makes
   the DOS version of the ghostscript to behave like a "ghostview")
   is available at
   ftp.muni.cz:pub/tex/local/postscript/psview.zip
     -- of from my web page too
   (anyway, the OS/2 and X-Windows ghostviews are much better. Upgrade your
   machine if you are sticked down by your TSR DOS).
   (this psview(.cmd) works also fine under OS/2, it could be well used
   for both editing the postscript and previewing immediately the result).

  In the beginning of 1999, the pm3d functionality has been implemented into
   gnuplot, the free scientific plotting program. There you can profit from
   pm3d display on OS/2 PM, X11, Linux VGA, Windows, Macintosh terminals
   directly (with support for mouse zoom etc.), with support for output to
   gif, png, postscript. You are encouraged to use it. See gnuplot.sf.net
   or www.gnuplot.info (or my web page) for more details.


  History:

    version 2.9.1: October 2003
    	- fixes for the new ANSI C++ compilers like GNU C++ 3.0: consequently,
	  no support for pipes like reading .gz files or writing to |gs
    version 2.9: February 2002
	- fixed typo !!B in EnlargeArrays()
    version 2.8: November 1999
	- small changes (buffered input and output, <<endl vs. <<'\n')
	- smoothing over n points if "take_n-th_point" is negative
    version 2.7: May 1999
	- input file generalization: input file can be stdin '-', file.gz
	  (automatically passed through gzip), or also any pipe, i.e. 
	  '"<awk -f ang2rec.awk angular.dat"' (notice that you must use 
	  quotes)
	  Notice that even the input .p3d file can be .gz-ipped
	- if the output file ends with .gz, then .ps.gz file is created by
	  automatically piping the output .ps via gzip
	- in the input file, input/output file names can contain spaces: just 
	  enclose them in quotes. Thus you can have an input from an awk 
	  script, for instance
    version 2.6b: April 1999
	- new colour formulae
    version 2.6a: March 1999
	- makefile for HP CC compiler
    version 2.6:  January 1999
	- some more output size optimization
	- added pm3d_map_begin / _end
	- the above changes propagated to pm3dCompress.awk
	- added 4 new colour formulae
	- December 1998 / January 1999: programming pm3d modes into
	  gnuplot. Maybe you can expect an interactive plotting to
	  various displays directly, output in .gif etc...
    version 2.5:  11.11.1998:
	- open in the output file in binary mode so that the end-of-line
	  chars are LF only even under OS/2 and DOS
	- number of entries in pmdict increased to 100
	- bugfix: tick width were not multiplied by pmlinewidth
	- draw tics after the whole map (important if they point into the map)
	- new feature: labels of Y ticks may be not only vertical, but
	  also horizontal
	- bounding box estimate improved
    version 2.4:  4.7.1998:
       - added paper size and orientation into the postscript header
       - bugfix for scan with 1 point only
    version 2.3:  5.4.1998:
       - ticks in and out
       - scale-invariant distances between axes and labels
       - bugfix of autoscale (was sometimes wrong in descending seq.)
    version 2.2:  31.10.1997:
       - colour maps
    version 2.1:  8.7.1997:
       - plotting of grid through ticks
       - minor ticks
       - bugfix for the lower key of the smooth box
    version 2.0: 16.3.1997:
       - Java GUI for pm3d (stand-alone front-end application Jpm3d)
       - new format of the input file (with labels, loadeable into Jpm3d)
	 (Jpm3d can save both formats, pm3d can read both formats,
	  Jpm3d can read only the new format, thus conversion between
	  them is not needed).
       - auto ticks on axes - big improvement
       - drawing the key box is done after the graph
    version 1.7: 20.8.1996:
       - {/Symbol _greek or symbol text_} is now acceptable anywhere in
	 the labels.
	 Drawback: centering nor flushing right does not work.
	 Solution (manual, but it works):
	 {/Symbol+0.4 _greek or symbol text_}  0.3 of
	 {/Symbol-0.2 _greek or symbol text_}
	   which shifts the whole string (not only /part with /Symbol, but
	   the whole text) +0.3 right (or up), or -0.3 left (or down), all
	   relative to the graph size
       - further, definition  /RotateShow { gsave 90 rotate show grestore } def
	 changed so that /Helvetica ... RotateShow /Symbol ... RotateShow other fonts
	 works as expected
       - parentheses (, ), { and } occuring in the text are preceded by backslash
       - ansireencoding ISO: Angstrom available (use char \305)
       - ticks on axes
       - minor problem with last scan drawing corrected
       - option -f (and the definition ScaleFont in the postscript file) added
	 This allows an easy way to increse the font size if you find it too small
    version 1.6: 17.7.1996:
       - flushes the scan left, right or center
    version 1.5: 21.1.1996:
       - pipes (into ghostscript) work fine now
       - truncation rods algorithm finished and documented
    version 1.4: 23.12.1995:
       - big smooth gray box intensity draw added
    version 1.3: 26.11.1995:
       - I've read a brief PostScript manual for the first time!
	 (so far I proceeded using the reverse engineering and
	 try and look methods). But in this version, I introduced
	 a lot of stuff in order to decrese the .ps file size, like:
       - only 3 points and fill command suffices to plot a quadrangle,
       - optional (-S command line param.) file size optimization.
	 Program checks the lines going to the .ps file whether their
	 line tails coincide -- if yes, than use abbreviations.
	 The old rapid version works if Speed=1, size is optimized
	 if Speed=0. Usually, if Size is optimized, than the file size
	 is decreased from 5% up to 50% with speed lost about 20%, and
	 GhostScript speed decrease about 7% (measured under OS/2 on my
	 computer).
       - I tried to plot the quadrangles relatively wrt the first one,
	 but it does not decrease the file size
       - and finally, the clipping was implemented
    version 1.2: 10.10.1995:
       - direct pipe (using e.g. -o "-o |gs") available. Tested under
	 OS/2 and gcc/emx
	 Not supported under DOS, since this OS is not multitasking.
       - output to stdout available
    version 1.1: 12.9.1995. What's new:
       - command line options
       - autoscale function available as a command line option
       - gray boxes with keys
    version 1.01, August 1995:
       - code of pm3d has not been changed
       - README a little bit changed
       - program MinMax added. This program calculates minima and maxima
	 of data arranged in columns in a data file. It can be used instead
	 of non-existing autoscale function within the pm3d program
    version 1: March/April 1995
    original version: my graphic program PMGRAF.EXE (for DOS)
      (Turbo Vision environment with editors and graphic tools
      for 2D and 3D graphics). To be found at ftp.muni.cz (see
      below) or my web page.


  Compiled and tested on:
    PC, Windows and GNU C++ (MingW and Cygwin)
    PC, OS/2  and GNU C++ (gcc/emx)
    PC, OS/2  and Borland C++ (bcos2; not working since 2.2 because of
		    the "Compiler stack overflow")
    PC, DOS   and GNU C++ (gcc/djgpp)
    PC, DOS   and BC++ 3.1 (bc)
    PC, Linux and GNU C++ (g++),
    Silicon Graphics and SGI C++ (CC)
       - compiler does not allow  (void*)out=(void*)&cout;
	 and there are more problems with it. I prefer to use g++
    HP-UX and GNU C/C++ (g++)
    HP-UX and HP C++ (CC)
    VMS and DEC CXX

  Enjoy!
*/

#if defined(__DECCXX) || defined(__unix)
  // #define stricmp strcmp
  #define stricmp strcasecmp
#endif

/*
   For future:
     - read each s-th scan (now it reads all scans, and every n-th point)
       Till it is implemented, use  awk  as the filter
*/


//                        ***  END OF NOTES  ***
///////////////////////////////////////////////////////////////////////



/* Notes for programmers:
   1. arrays Ax,Ay,Az keep values of the previous scan,
      arrays Bx,By,Bz keep values of the current scan
      read from Inp file
   2. contrary to my original Pascal version, aimax and
      bimax are numbers of points in each scan
*/

#ifdef __TURBOC__
  #include <conio.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <fstream.h>
#include <iomanip.h>
#include "ansicpp.h"
#include "getoptpm.h"
#include "lfstream.h"
#include "pmparser.h"
#include "pmruzne.h"
#include "theline.h"

#define TheDate "October 2003"


#define MaxOutProblems 10
#ifdef CERR_FLUSH
  #undef CERR_FLUSH
#endif
#define CERR_FLUSH COUT_FLUSH


/* file descriptors not supported by the latest ANSI C++ standard
 * => no pipes possible in GNU C++ 3.0 and other compilers, sorry!
 */
#if (defined(__GNUC__) && __GNUC__>=3)
  #ifdef PIPES
    #undef PIPES
  #endif
#endif


/*
// This is obsolete! I found that all this is in  fileno(FILE*) function!
// Now define file descriptor or file handle. This is the integer number
// in the FILE structure defined somewhere in stdio.h. It is the unique
// number for each open file and it allows the program to use both
// C and C++ access. Look into stdio.h for its name if systems or compilers
// not listed below.
#if   defined(__EMX__)
  #if __GNUC_MINOR__==6   // OK for gcc 2.6.3
     #define FILE_HANDLE handle
  #else
     #define FILE_HANDLE _handle  // OK for gcc 2.7.2
  #endif
#elif defined(__BORLANDC__)
  #define FILE_HANDLE fd
#elif defined(__linux)
  #define FILE_HANDLE _fileno
#else
  // Unix: like below, or open /dev/fdn, where n is handle number
  #define FILE_HANDLE _file
#endif
*/


/*
This could be defined in order to plot all the quadrangles
relatively wrt the first one. But it makes no improvement
in the .ps file size, thus I let it undefined.
*/
// #define RelativeRectangles


// #define gs261  // there were other values for paper sizes?
// Or had I used another output device?

// ** Basic function **

inline int MAX ( int X, int Y )
{
return (X>Y) ? X:Y;
}

inline int MAX ( word X, word Y )
{
return (X>Y) ? X:Y;
}

inline double MAX ( double X, double Y )
{
return (X>Y) ? X:Y;
}

inline word MIN ( int X, int Y )
{
return (X<Y) ? X:Y;
}

inline word MIN ( word X, word Y )
{
return (X<Y) ? X:Y;
}

inline double MIN ( double X, double Y )
{
return (X<Y) ? X:Y;
}

inline int iround ( double x ) { return (int)floor(x+0.5); }

#define sqr(x) ((x)*(x))

const double ln_10   = 2.302585092994;
inline double alog10 ( double x )
{
return exp(ln_10*x);
}


/* // this routine works, but I need the next one
char* memcmp_rev ( char *S1, char *S2, int maxlen ) {
// reverse compare of S1 and S2 up to maxlen chars.
// returns the last common char
if (maxlen<=0) return S2;
while (--maxlen>0 && *(--S1)==*(--S2))
 { cout <<"S1="<<S1<<"   S2="<<S2<<endl; }
if (maxlen<=0) return S2;
return ++S2;
}
*/


char* memcmp_rev_space ( char *S1, char *S2, int maxlen )
{
// reverse word compare of S1 and S2 up to maxlen chars.
// returns the last common char, which is space (word separator)
if (maxlen<=0) return S2;
while (--maxlen>0 && *(--S1)==*(--S2));
if (maxlen<=0) return S2;
// *S2 points to the char that is different in both string
while (*(++S2) && *S2!=' '); //find space
return S2;
}


word ArrayMax = 1024;       // allocated nb. of points in scan
#define ArrayDelta 1024     // step for the enlargement
double *Ax,*Ay,*Az, *Bx,*By,*Bz;
char *Ain,*Bin;   // 1 if the point Ax[i],Ay[i] is inside the graph region
		  // is one can plot the correspoding quadrangle
double Zinterval; // Zinterval = Zto-Zfrom;
double Color, MaxColor, StartColor; // for gray map: levels of gray
double ScaleX,ScaleY,  // determines the graph size, see routines x2p, y2p
       FontScale;      // determines the font size
word Scans, MaxAimax;  // counters for the final information
char Title[128]="",
     LabelX[128]="",
     LabelY[128]="";
char Flush = 'L';
double BoundBoxXA,BoundBoxYA, // boundary box (approximate, not any problem
       BoundBoxXB,BoundBoxYB; // so far). How to get it exactly?
lfstream Inp; // input file (derived from class lfstream from lfstream.h)
ostream *out; // output postscript file
#ifdef PIPES
FILE *PipeLikeFile; // needed to open an output pipe
#endif
char CanCloseOut, // file must be closed, cout cannot
     WaitKillOut=0;     // kill this process by Ctrl/C
int outproblems=0; // exit if too much output problems
double PaperX,PaperY,   // values dependent on page size (they depend on
       ShiftX0,ShiftY0; // output device, density or whatever). I have found
       // these values for dos,dosl, a4,a4l, see the routine PaperName2Values
char XticksOn=0, YticksOn=0;  // were the above given on the cmd line?
double XticksFrom=-1e305, XticksStep=-1, XticksTo=-2e305;
double YticksFrom=-1e305, YticksStep=-1, YticksTo=-2e305;
double MinorTicksX=1, MinorTicksY=1; // minor ticks frequency
int WhereMinorX=0, WhereMinorY=0;    // 0 (no), 1 (one side), 2 (both sides)
#define gridNever  0
#define gridBefore 1
#define gridAfter  2
char whenGridX=gridNever, whenGridY=gridNever;
char MajorGridX=1, MajorGridY=1; // if do grid, then in which ticks
char DoAutoscaleX=0, DoAutoscaleY=0, DoAutoscaleZ=0;
char TicksOut = 1;
int SmoothEveryNth = 0; // 0 for unused

// definition of the colour formulae
const int colorFormulae = 34;
char (ColorFormula[2*colorFormulae])[90] = {
  "pop 0",		"0",
  "pop 0.5",		"0.5",
  "pop 1",		"1",
  "",			"x",
  "dup mul",		"x^2",
  "dup dup mul mul",	"x^3",
  "dup mul dup mul",	"x^4",
  "sqrt",		"sqrt(x)",
  "sqrt sqrt",		"sqrt(sqrt(x))",
  "90 mul sin",		"sin(90x)",
  "90 mul cos",		"cos(90x)",
  "0.5 sub abs",	"|x-0.5|",
  "2 mul 1 sub dup mul","(2x-1)^2",
  "180 mul sin",	"sin(180x)",
  "180 mul cos abs",	"|cos(180x)|",
  "360 mul sin",	"sin(360x)",
  "360 mul cos",	"cos(360x)",
  "360 mul sin abs",	"|sin(360x)|",
  "360 mul cos abs",	"|cos(360x)|",
  "720 mul sin abs",	"|sin(720x)|",
  "720 mul cos abs",	"|cos(720x)|",
  "3 mul",		"3x",
  "3 mul 1 sub",	"3x-1",
  "3 mul 2 sub",	"3x-2",
  "3 mul 1 sub abs",	"|3x-1|",
  "3 mul 2 sub abs",	"|3x-2|",
  "1.5 mul .5 sub",	"(3x-1)/2",
  "1.5 mul 1 sub",	"(3x-2)/2",
  "1.5 mul .5 sub abs",	"|(3x-1)/2|",
  "1.5 mul 1 sub abs",	"|(3x-2)/2|",
  "0.32 div 0.78125 sub", "x/0.32-0.78125",
  "2 mul 0.84 sub",	"2*x-0.84",
  "dup 0.42 le {4 mul} {dup 0.92 le {-2 mul 1.84 add} {0.08 div 11.5 sub} ifelse} ifelse", "4x;1;-2x+1.84;x/0.08-11.5",
  "2 mul 0.5 sub abs",	"|2*x - 0.5|"
  };

/**  supplied variables, which are read from             **/
/**  the input file  pm3d.p3d  or from the command line  **/
char *InpFile=0, *OutFile=0;   // file names

#define colorModeGRAY 'g'
#define colorModeRGB  'r'
#define availableCOLORMODES "gr"
char ColorMode='g';
char *GrayFormula=0;              // gray transformation expression
int formulaR=7, formulaB=15, formulaG=5;

long FromLine=3, FromScan=1;         // that's clear
char   readnotes=1;                  // notes from the first 2 lines of the
double NotesX=0, NotesY=-0.15;       // data file
double Xfrom=0,Xto=10, Yfrom=0,Yto=10, // limits. They can be changed by the
       Zfrom=0,Zto=10;                 // -a param  autoscale option
int  isZlinlog=0,      // 0/1 Z is linear/logarithmic
     landscape=1,      // landscape mode?
     isboundary=1;     // draw the boundary around the graph?
char isnegative=0,     // negative image
     isbackblack=0;    // file the regions by black?
double GRSizeX=0.8,GRSizeY=0.8,  // relative size (wrt graph size)
       ShiftX=0,ShiftY=0;        // relative shift
int    ZkeyIntervals=4;             // number of gray boxes
double RelRefX=1.03, RelRefY=0.8,   // and their relative position
       RelBoxSizeX=0.05, RelBoxSizeY=0.05;  // and rel. size of the smooth box
double ScaleFont=1.0; // relative font size
char VerticalYTicksLabels=1; // default: vertical labes of Y ticks (otherwise horizontal)

// In pmgraf.exe, I can draw a 3d graph not only from a file, but
// from an analytical function. This is not available here,
// so don't worry about ifFF and isFFand
// char isFromWhich;
// #define isFromFile 0
// not like that, but (FF means isFromFile):
#define ifFF {}
#define isFFand

/* If one would have a colour map with 16 colours,
   this colour sequence is pleasant (from pmgraf.exe):
   Black, DarkGray, Brown, LightGray, White, LightCyan, Cyan,
   LightGreen, Green, LightMagenta, Magenta,
   Blue, LightBlue, Yellow, LightRed, Red
*/

// buffered output
#ifdef __MSDOS__
  const int size_input_buffer = 64*1024;
  const int size_output_buffer = 64*1024;
#else
  const int size_input_buffer = 2*1024;
  const int size_output_buffer = 2*1024;
#endif
char input_buffer[ size_input_buffer ];
char output_buffer[ size_output_buffer ];


// Now some general routines

inline double x2p ( double x )
{
// transformation from x coordinate to x paper coordinate
// A unit of paper coordinate is 0.01 bp
return ShiftX + ScaleX * (x-Xfrom);
}


inline double y2p ( double y )
{
// transformation from y coordinate to y paper coordinate
return ShiftY + ScaleY * (y-Yfrom);
}


inline double wrtGraphX ( double x )
{
return ShiftX+x*GRSizeX*PaperX;
}


inline double wrtGraphY ( double y )
{
return ShiftY+y*GRSizeY*PaperY;
}


double GetFontScale ( double scale )
{
return iround(MIN(GRSizeX,GRSizeY)*scale/FontScale);
}


void font ( double scale )
{
// entitle a string printing
*out <<"/Helvetica findfont "<<GetFontScale(scale)
     <<" ScaleFont setfont ";
}


void WriteTail ()
{
// the string just before the end of the output file
*out << "grestore end\n";
}


void WriteTail2 ()
{
// the string just before the end of th output file
*out << "showpage\n%eof\n";
}


char TextBuffer[2048]; // I hope nobody will have longer comment

void Expanded_strcpy ( char **TextBufferUp, char *TitleUp )
{
// string with (){} preceded by '\'
while (*TitleUp) {
  if (strchr("(){}",*TitleUp)) *(*TextBufferUp)++ = '\\';
  *(*TextBufferUp)++ = *TitleUp++; }
**TextBufferUp = *TitleUp;
}


void Expanded_memcpy ( char **TextBufferUp, char *TitleUp, size_t n )
{
// string with (){} preceded by '\'
if (!n) **TextBufferUp = 0; else {
while (n--) {
  if (strchr("(){}",*TitleUp)) *(*TextBufferUp)++ = '\\';
  *(*TextBufferUp)++ = *TitleUp++;
  }
}
}


char ExpandText ( int scale, char *Title, char *TextBuffer, char *show,
	double& shift )
{
// show should be "show" or "RotateShow"
// returns 1 if the text has been expanded

int i,counter,t,itmp;
char *cLeft, *cRight, *c;
char *TitleUp = Title;
char *TextBufferUp = TextBuffer; *TextBuffer=0;
char res=0;

shift = 0;
scale = iround(MIN(GRSizeX,GRSizeY)*scale/FontScale);

do {
cLeft = strstr(TitleUp,"{/Symbol");
if (!cLeft) {
  Expanded_strcpy(&TextBufferUp,TitleUp);
  return res; }

// find shift (if given)
c = cLeft+8;
if (*c=='+' || *c=='-') {
  shift=strtod(c,&cRight); // read that number
  // cut the number, i.e. the string between c and cRight-1
  while ((*c++ = *cRight++));
  }

// find corresponding }
counter = 1;
for (cRight=cLeft+9; *cRight && counter; cRight++)
  if (*cRight=='}') counter--;
    else if (*cRight=='{') counter++;

if (*cRight==0 && counter>0) // { and } do not match
   { Expanded_strcpy(&TitleUp,TextBufferUp); return res; }

res=1;

t=cLeft-TitleUp-1; // copy all up to {/Symbol
Expanded_memcpy(&TextBufferUp,Title,t+1);
// TextBufferUp[++t]=0;
*TextBufferUp=0;

// cout<<"TB="<<TextBuffer<<"=\n";
sprintf(TextBufferUp+strlen(TextBufferUp),
  ") %s /Symbol findfont %d ScaleFont setfont (",show,int(scale));

itmp=strlen(TextBufferUp);
TextBufferUp += itmp;

Expanded_memcpy(&TextBufferUp,cLeft+9,cRight-cLeft-10);
// TextBufferUp-=itmp;
*TextBufferUp = 0;
if (!*cRight) return 1;

sprintf(TextBufferUp+strlen(TextBufferUp),
  ") %s /Helvetica findfont %d ScaleFont setfont (",show,int(scale));

// copy all up to next Symbol character
TitleUp = cRight;
cLeft = strstr(TitleUp,"{/Symbol ");
if (!cLeft) cLeft = TitleUp+strlen(TitleUp);
i = cLeft-cRight;
t = strlen(TextBufferUp);
TextBufferUp += t;
Expanded_memcpy(&TextBufferUp,cRight,i);
*TextBufferUp = 0;
if (!*cLeft) return 1;

// TextBufferUp = TextBuffer + strlen(TextBuffer)-t;
TitleUp = cLeft;
} while (1);

}


/*
void pmemcpy ( void *dest, void *src, long bytes ) {
// like memcpy, but for long arrays
#define N 32535
while (bytes>N) {
  memcpy(dest,src,N); (char*)src+=N;
  (char*)dest+=N; bytes-=N; }
memcpy(dest,src,int(bytes));
#undef N
}
*/
#define pmemcpy memcpy

void SetColor ( double color )
{
static double OldColor = 1511.1969;
if (color<0) OldColor = 1511.1969; /* clear saved value*/ else
if (color==OldColor) return;
OldColor = color;
#if 0
*out << setprecision(4) << color << " g ";
#else
// round to 4 digits, but don't write the leading 0 before the decimal point
if (color<=0) *out <<"0 g "; else
if (color>=1) *out <<"1 g "; else {
   char s[15];
   color = 0.0001*(int)(color*10000+0.5); // round it
   sprintf(s,"%.4g g ",color);
   if (s[0]=='0' && s[1]=='.') *out << &(s[1]);
     else *out << s;
  }
#endif
}


double Get3dColor ( double Z )
{
// this transforms the z coordinate to gray (not yet far: or to colour)
double color;
if (!GrayFormula) {
    if (Z<Zfrom) color = StartColor;
      else
	if (Z>Zto) color = MaxColor;
	  else color = StartColor + (Z-Zfrom)/Zinterval;
    }
  else // is gray transformation
    if (GrayFormula) {
      static double Dollars[2];
      if (Z<Zfrom) Dollars[1] = 0; else if (Z>Zto) Dollars[1] = 1;
	else Dollars[1] = (Z-Zfrom)/Zinterval;
      color = FormulaParserDollars(GrayFormula,Dollars,1);
      if (color<0) color = 0;
	else if (color>1) color = 1;
      }
if (isnegative) color = 1 - color;
return color;
}


inline double Get3dColori ( word ai, word bi )
{
// gray (or colour) is given according to the arithmetic mean
// of the Z values of the 4 surrounding points
return Get3dColor( ( Az[ai]+Az[ai+1]+Bz[bi]+Bz[bi+1] ) * 0.25 );
}


void WriteHeader () {
// write down the beginning of the (postscript) output file
*out<<"%!PS-Adobe-2.0\n\
%%Creator: Program  pm3d  by Petr Mikulik\n\
%%BoundingBox: "<<iround(BoundBoxXA*0.01)<<' '<<iround(BoundBoxYA*0.01)<<' ';
if (!landscape) *out<<iround(BoundBoxXB*0.01)<<' '<<iround(BoundBoxYB*0.01)<<'\n';
  else  *out<<iround(BoundBoxYB*0.01)<<' '<<iround(BoundBoxXB*0.01)<<endl;
*out<<"%%Pages: 1\n\
%%Orientation: ";
 if (landscape) *out<<"Landscape"; else *out<<"Portrait";
*out<<"\n%%DocumentPaperSizes: A4\n\
%%DocumentFonts: Helvetica\n\
%%EndComments\n";
*out<<"%pm3dCommandLine: pm3d";
int i, k;
for (i=1; i<ansicpp_argc; ) *out << " " << ansicpp_argv[i++]; *out<<'\n';
*out<<"/pmdict 100 dict def\n\
pmdict begin\n\
/pmlinewidth 10 def\n\
/B {bind def} def\n\
/ScaleFont {"<<ScaleFont<<" mul scalefont}B\t\t%adjust your font size here\n\
/vshift -46 def\n\
/D {def} def\n\
/M {moveto}B\n\
/L {lineto}B\n\
/R {rmoveto}B\n\
/V {rlineto}B\n\
/v {moveto rlineto stroke}B\n\
/G {setgray}B\n";

for (i=0; i<colorFormulae; i++) {
  *out<<"/cF"<<i<<" {"<<ColorFormula[2*i]<<"}B";
  k = 15 + strlen(ColorFormula[2*i]);
  k += (i<10) ? 1:2;
  if (k<24) *out<<"\t\t"; else if (k<32) *out<<'\t';
  *out<<"\t% "<<ColorFormula[2*i+1]<<'\n';
  }
/* // this was in the first colour version:
/cF1 {sqrt}B\n\
/cF2 {dup dup mul mul}B\n\
/cF3 {360 mul sin}B\n\
*/
if (abs(formulaR)>=colorFormulae) formulaR=0;
if (abs(formulaG)>=colorFormulae) formulaG=0;
if (abs(formulaB)>=colorFormulae) formulaB=0;
*out<<"/g {dup ";
if (formulaR<0) *out<<"1 exch sub "; // negate
*out<<"cF"<<abs(formulaR)<<" exch dup ";
if (formulaG<0) *out<<"1 exch sub "; // negate
*out<<"cF"<<abs(formulaG)<<" exch\n\t";
if (formulaB<0) *out<<"1 exch sub "; // negate
*out<<"cF"<<abs(formulaB)<<" setrgbcolor}B\t% uncommented for COLOUR RGB map\n";
if (ColorMode!=colorModeGRAY) *out<<"% ";
*out<<"/g {setgray}B\t\t\t\t% uncommented for GRAY map\n";

*out<<"/N {newpath moveto} def\n\
/C {closepath} def\n\
/F {fill} def\n\
/s {rlineto closepath gsave fill grestore 0 setgray stroke}B\n\
/f {rlineto fill}B\n\
/h {rlineto rlineto rlineto fill}B\n\
/w {setlinewidth}B\n\
/X {closepath fill stroke} def\n\
/topshift {(m) stringwidth pop neg} def\n\
/showTopAlign {0 topshift R show} def\n\
/Rshow { currentpoint stroke M\n\
  dup stringwidth pop neg vshift R show } def\n\
/RshowTopAlign { currentpoint stroke M\n\
  dup stringwidth pop neg topshift R show } def\n\
/RshowMidAlign { currentpoint stroke M\n\
  dup stringwidth pop neg topshift 2 div R show } def\n\
/Cshow { currentpoint stroke M\n\
  dup stringwidth pop -2 div vshift R show } def\n\
/CshowTopAlign { currentpoint stroke M\n\
  dup stringwidth pop -2 div topshift R show } def\n\
/RotateShow { 90 rotate show -90 rotate } def\n\
/RotateRshow { gsave 90 rotate currentpoint stroke M\n\
  dup stringwidth pop neg vshift R show grestore } def\n\
/RotateCshow { gsave 90 rotate currentpoint stroke M\n\
  dup stringwidth pop -2 div vshift R show grestore } def\
\n\
/reencodeISO {\n\
dup dup findfont dup length dict begin\n\
{ 1 index /FID ne { def }{ pop pop } ifelse } forall\n\
/Encoding ISOLatin1Encoding def\n\
currentdict end definefont\n\
} def\n\
/ISOLatin1Encoding [\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef\n\
/space/exclam/quotedbl/numbersign/dollar/percent/ampersand/quoteright\n\
/parenleft/parenright/asterisk/plus/comma/minus/period/slash\n\
/zero/one/two/three/four/five/six/seven/eight/nine/colon/semicolon\n\
/less/equal/greater/question/at/A/B/C/D/E/F/G/H/I/J/K/L/M/N\n\
/O/P/Q/R/S/T/U/V/W/X/Y/Z/bracketleft/backslash/bracketright\n\
/asciicircum/underscore/quoteleft/a/b/c/d/e/f/g/h/i/j/k/l/m\n\
/n/o/p/q/r/s/t/u/v/w/x/y/z/braceleft/bar/braceright/asciitilde\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef\n\
/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef/.notdef\n\
/.notdef/dotlessi/grave/acute/circumflex/tilde/macron/breve\n\
/dotaccent/dieresis/.notdef/ring/cedilla/.notdef/hungarumlaut\n\
/ogonek/caron/space/exclamdown/cent/sterling/currency/yen/brokenbar\n\
/section/dieresis/copyright/ordfeminine/guillemotleft/logicalnot\n\
/hyphen/registered/macron/degree/plusminus/twosuperior/threesuperior\n\
/acute/mu/paragraph/periodcentered/cedilla/onesuperior/ordmasculine\n\
/guillemotright/onequarter/onehalf/threequarters/questiondown\n\
/Agrave/Aacute/Acircumflex/Atilde/Adieresis/Aring/AE/Ccedilla\n\
/Egrave/Eacute/Ecircumflex/Edieresis/Igrave/Iacute/Icircumflex\n\
/Idieresis/Eth/Ntilde/Ograve/Oacute/Ocircumflex/Otilde/Odieresis\n\
/multiply/Oslash/Ugrave/Uacute/Ucircumflex/Udieresis/Yacute\n\
/Thorn/germandbls/agrave/aacute/acircumflex/atilde/adieresis\n\
/aring/ae/ccedilla/egrave/eacute/ecircumflex/edieresis/igrave\n\
/iacute/icircumflex/idieresis/eth/ntilde/ograve/oacute/ocircumflex\n\
/otilde/odieresis/divide/oslash/ugrave/uacute/ucircumflex/udieresis\n\
/yacute/thorn/ydieresis\n\
] def\n\
/Helvetica reencodeISO def\n";
#ifdef RelativeRectangles
*out<<"/n {newpath moveto currentpoint} def\n";
#endif
*out<<"end\n\
%%EndProlog\n\n\
%%Page: 1 1\n\
pmdict begin\n\
gsave\n";
*out << "%0 0 translate\n";
*out << "0.01 0.01 scale\t\t\t\t\t\t%the unit is 0.01 bp\n";
if (!landscape)
  *out <<"\t\t\t%uncomment the line below for landscape mode\n%";
*out <<PaperY<<" 0 translate 90 rotate\t\t\t\t%landscape mode\n";
// Seascape: *out <<"0 "<<PaperX<<" translate -90 rotate\t\t\t\t%landscape mode\n";
double xA,yA,xB,yB;
if (isbackblack) { // draw black rectangle
  *out <<"0 w\t\t\t\t%black rectangle\n";
  xA=floor(x2p(Xfrom)); yA=floor(y2p(Yfrom));
  xB=floor(x2p(Xto));   yB=floor(y2p(Yto));
  *out << "0 g "<< xA <<' '<< yA <<" N ";
  *out << xB <<' '<< yA <<" L ";
  *out << xB <<' '<< yB <<" L ";
  *out << xA <<' '<< yB <<" L ";
  *out << xA <<' '<< yA <<" L X\n";
  }
}


void Xticks ( )
{
double x;
// find ticks if not given
if (!XticksOn) { XticksFrom=Xfrom; XticksTo=Xto; }
if (!XticksOn || XticksStep<=0) { // adjust the step on the axis
  XticksStep=(XticksTo-XticksFrom)/9; // we want 10 intervals by default
  x = alog10( (int)(log10(XticksStep)) );
  if (XticksStep<1) x/=10;
  XticksStep = (int)(XticksStep/x+0.0001);  // I let only the first digit, i.e. 0<step<10
  // now avoid Steps being 3,4,6,7,8,9:
  if (XticksStep>2.5 && XticksStep<6.5) XticksStep=5;
    else if (XticksStep>=6.5) XticksStep=10;
  XticksStep *= x;
  }
if (XticksOn && XticksFrom<Xfrom) {
  XticksOn=0; XticksFrom=Xfrom; }
if (!XticksOn) { // round them wrt step
  XticksFrom=XticksStep*ceil(XticksFrom/XticksStep-0.00001);
  XticksTo=XticksStep*floor(XticksTo/XticksStep+0.00001);
  }
if (XticksTo<XticksFrom) {
  *out <<"/MakeGridX {} def\n"; return; }
//cerr<<"Xticks From="<<XticksFrom<<endl;cerr<<"Xticks Step="<<XticksStep<<endl;
//cerr<<"Xticks To  ="<<XticksTo  <<endl; exit(0);
// now, plot them
*out<<"/TicsX {\t\t% define ticks on x axis\n";
*out<<"0 G 5 pmlinewidth w "; font(4500); *out<<'\n';
font(3000); *out<<'\n';
// sprintf(LabelX,"%1.8g",Xfrom);
long ldy = ( (long) (wrtGraphY(0.02)-wrtGraphY(0))    // tick length
	  + (long) (wrtGraphX(0.02)-wrtGraphX(0)) )/2;
long ldy2 = (long) (1*ldy);                       // space between tick and topline of the number
long ly = long( floor(y2p(Yfrom)) ),
     lyTo = long( floor(y2p(Yto)) ),
     lx;
x = XticksFrom; // older: = MAX(XticksFrom,Xfrom);
double XticksUp = MIN(XticksTo,Xto)+XticksStep*1e-6;
int tickcounter=0;
for ( ; x<=XticksUp; x=XticksFrom+(++tickcounter)*XticksStep ) {
  if (fabs(x)<1e-14*XticksStep) x=0;
  lx=long( floor(x2p(x)) );
  // sprintf(LabelX,"%1.8g",Xto);
  *out<<lx<<" "<<ly<<" M 0 ";
  if (TicksOut) *out<<-ldy<<" V 0 "<<-ldy2;
    else *out<<ldy<<" V 0 "<<-2*ldy2;
  *out<<" R ("<<x<<") ";
  if (0 && !tickcounter) *out<<"showTopAlign\n";  // x axis can be from negative number, not nice
    else *out<<"CshowTopAlign\n"; // 4.4.1998
    // else if (XticksOn) *out<<"CshowTopAlign\n"; else *out<<"RshowTopAlign\n";
    // the first tick, or both if no ticks specified on the command line
  if (WhereMinorX==2) { // tics on the other side 9.11.1998
    *out<<'\t'<<lx<<" "<<lyTo<<" M 0 ";
    if (TicksOut) *out<<ldy<<" V\n";
      else *out<<-ldy<<" V\n";
    }
  }
*out<<"stroke} def\n";
// Now draw grid in x
double TheMinorTicksX=MinorTicksX;
if (MajorGridX || TheMinorTicksX<0.005) TheMinorTicksX=1;
long lxstep = long( (x2p(XticksStep)-x2p(0))/TheMinorTicksX ),
     ylen = long(wrtGraphY(1)-wrtGraphY(0));
*out<<"/DrawGridX {\t\t\t\t\t\t\t% define grid x\n"
    <<"gsave pmlinewidth w 0 G\n";
*out<<"GridStartPosMain GridStartPos M ";
*out<<"/ii 0 def % nb of grid lines, counter\n";
*out<<"{GridStep 0 R gsave 0 GridLineLen V stroke grestore\n";
*out<< "/ii ii 1 add def ii imax gt {exit} if } loop\n"
    << "grestore } def % end of grid x\n";
*out<<"/MakeGridX { /GridStartPos "<<ly
    <<" def/GridStartPosMain "<<long( x2p(XticksFrom-XticksStep/TheMinorTicksX) )
    <<"\ndef /GridLineLen "<<ylen
    <<" def /GridStep "<<lxstep
    <<" def /imax "<<long((tickcounter-1)*TheMinorTicksX)<<" def DrawGridX } def\n";
// minor ticks here
long ticklen = ldy/2; // (long) (wrtGraphY(0.01)-wrtGraphY(0))/2; // tick length
if (!TicksOut) ticklen=-ticklen;
lxstep = long( (x2p(XticksStep)-x2p(0))/MinorTicksX );
*out<<"/GridStartPosMain "<<long( x2p(XticksFrom-XticksStep/MinorTicksX) )
    <<" def /GridStep "<<lxstep
    <<" def /imax "<<long((tickcounter-1)*MinorTicksX)
    <<" def\t% minor ticks x\n";if (WhereMinorX==0) *out<<"%"; // no minor ticks
*out<<"/GridStartPos "<<ly
    <<" def /GridLineLen "<<-ticklen	// one side
    <<" def DrawGridX\t\t% down\n";
if (WhereMinorX!=2) *out<<"%"; // no these minor ticks
*out<<"/GridStartPos "<<long( floor(y2p(Yto)) )
    <<" def /GridLineLen "<<ticklen	// one side
    <<" def DrawGridX\t\t% up\n";
}


void Yticks ( )
{
double y;
// find ticks if not given
if (!YticksOn) { YticksFrom=Yfrom; YticksTo=Yto; }
if (!YticksOn || YticksStep<=0) { // adjust the step on the axis
  YticksStep=(YticksTo-YticksFrom)/9; // we want 10 intervals by default
  y = alog10( (int)(log10(YticksStep)) );
  if (YticksStep<1) y/=10;
  YticksStep = (int)(YticksStep/y+0.0001);  // I let only the first digit, i.e. 0<step<10
  // now avoid Steps being 3,4,6,7,8,9:
  if (YticksStep>2.5 && YticksStep<6.5) YticksStep=5;
    else if (YticksStep>=6.5) YticksStep=10;
  YticksStep *= y;
  }
if (YticksOn && YticksFrom<Yfrom) {
  YticksOn=0; YticksFrom=Yfrom; }
if (!YticksOn) { // round them wrt step
  YticksFrom=YticksStep*ceil(YticksFrom/YticksStep-0.00001);
  YticksTo=YticksStep*floor(YticksTo/YticksStep+0.00001);
  }
if (YticksTo<YticksFrom) {
  *out <<"/MakeGridY {} def\n"; return; }
// now, plot them
*out<<"/TicsY {\t\t% define ticks on y axis\n";
*out<<"0 G 5 pmlinewidth w "; font(4500); *out<<'\n';
font(3000); *out<<'\n';
// sprintf(LabelX,"%1.8g",Xfrom);
long ldx = (long) ((ScaleX/ScaleY)*(
   ( (long) (wrtGraphY(0.02)-wrtGraphY(0))    // tick length
	  + (long) (wrtGraphX(0.02)-wrtGraphX(0)) )/2
   ));
ldx = ((long) (wrtGraphY(0.02)-wrtGraphY(0))    // tick length
	  + (long) (wrtGraphX(0.02)-wrtGraphX(0)) )/2;
long ldx2 = (long) (1*ldx);                     // space between tick and baseline of the number
long lx  = long( floor(x2p(Xfrom)) ),
     lxTo = long( floor(x2p(Xto)) ),
     ly;
y = YticksFrom; // older: = MAX(YticksFrom,Yfrom);
double YticksUp = MIN(YticksTo,Yto)+YticksStep*1e-6;
int tickcounter=0;
for ( ; y<=YticksUp; y=YticksFrom+(++tickcounter)*YticksStep ) {
  if (fabs(y)<1e-14*YticksStep) y=0;
  ly=long( floor(y2p(y)) );
  *out<<lx<<" "<<ly<<" M ";
  if (TicksOut) *out<<-ldx<<" 0 V "<<-ldx2;
    else *out<<ldx<<" 0 V "<<-2*ldx2;
  *out<<" 0 R ("<<y<<") ";
  if (0) {} // (!tickcounter) *out<<"RotateShow\n";
    else {
	if (VerticalYTicksLabels)
	    *out<<"RotateCshow\n"; // 4.4.1998
	  else
	    *out<<"RshowMidAlign\n"; // 10.11.1998
	}
    //else if (YticksOn) *out<<"RotateCshow\n"; else *out<<"RotateRshow\n";
  if (WhereMinorY==2) { // tics on the other side 9.11.1998
    *out<<'\t'<<lxTo<<" "<<ly<<" M ";
    if (TicksOut) *out<<ldx<<" 0 V\n";
      else *out<<-ldx<<" 0 V\n";
    }
  }
*out<<"stroke} def\n";
// Now draw grid in y
double TheMinorTicksY=MinorTicksY;
if (MajorGridY || MinorTicksY<0.005) TheMinorTicksY=1;
long lystep = long( (y2p(YticksStep)-y2p(0))/TheMinorTicksY ),
     xlen = long(wrtGraphX(1)-wrtGraphX(0));
*out<<"/DrawGridY {\t\t\t\t\t\t\t% define grid y\n"
    <<"gsave pmlinewidth w 0 G\n";
*out<<"GridStartPos GridStartPosMain M ";
*out<<"/ii 0 def % nb of grid lines, counter\n";
*out<<"{0 GridStep R gsave GridLineLen 0 V stroke grestore\n";
*out<< "/ii ii 1 add def ii imax gt {exit} if } loop\n"
    << "grestore } def % end of grid y\n";
*out<<"/MakeGridY {/GridStartPos "<<lx
    <<" def/GridStartPosMain "<<long( y2p(YticksFrom-YticksStep/TheMinorTicksY) )
    <<"\ndef /GridLineLen "<<xlen
    <<" def /GridStep "<<lystep
    <<" def /imax "<<long((tickcounter-1)*TheMinorTicksY)<<" def DrawGridY } def\n";
// minor ticks here
long ticklen = ldx/2; // (long) (wrtGraphX(0.01)-wrtGraphX(0))/2; // tick length
if (!TicksOut) ticklen=-ticklen;
lystep = long( (y2p(YticksStep)-y2p(0))/MinorTicksY );
*out<<"/GridStartPosMain "<<long( y2p(YticksFrom-YticksStep/MinorTicksY) )
    <<" def /GridStep "<<lystep
    <<" def /imax "<<long((tickcounter-1)*MinorTicksY)
    <<" def\t% minor ticks y\n";
if (WhereMinorY==0) *out<<"%"; // no minor ticks
*out<<"/GridStartPos "<<lx    <<" def /GridLineLen "<<-ticklen	// one side
    <<" def DrawGridY\t\t% left\n";
if (WhereMinorY!=2) *out<<"%"; // no these minor ticks
*out<<"/GridStartPos "<<long( floor(x2p(Xto)) )
    <<" def /GridLineLen "<<ticklen	// one side
    <<" def DrawGridY\t\t% right\n";
}


void MakeGridXY()
{
static char before=1;
if ((before && whenGridX!=gridBefore) || (!before && whenGridX!=gridAfter))
  *out <<"% ";
*out << "MakeGridX\t\t\t\t\t\t\t% draw grid X\n";
if ((before && whenGridY!=gridBefore) || (!before && whenGridY!=gridAfter))
  *out <<"% ";
*out << "MakeGridY\t\t\t\t\t\t\t% draw grid Y\n";
before=0;
}


void DoKeyOfDraw3D_smooth_box ()
{
char charTmp[80];
// make intensity smooth box (from black to white)
*out <<"\n% Draw Gray Scale Smooth Box\ngsave\n";
*out <<"/imax 1024 def %nb. of discrete steps; increase it if strips are visible\n";
double xA=long( wrtGraphX(RelRefX) ), yA=long( wrtGraphY(RelRefY) );
double SizeX=long( RelBoxSizeX*GRSizeX*PaperX ),
       SizeY=long( RelBoxSizeY*GRSizeY*PaperY );
*out<<xA<<' '<<yA<<" translate "<<SizeX<<' '<<SizeY<<" scale % def. left bottom corner\n";
*out<<"% relative size of the box. Then all coordinates of box are from [0,0] up to [1,1]\n";
*out <<"/ystep 1 imax div def %Now local variables; this one is step\n";
*out <<"0 w /y0 0 def /ii 0 def { %counters\n";
if (isnegative) *out <<"1 y0 sub g % \"y0 g\"  for positive,\"1 y0 sub g\"  for negative\n";
  else *out <<"y0 g % \"y0 g\"  for positive,\"1 y0 sub g\"  for negative\n";
*out <<"0 y0 N 1 0 V 0 ystep V -1 0 f\n";
*out <<"/y0 y0 ystep add def /ii ii 1 add def\n";
*out <<"ii imax gt {exit} if } loop\n";
*out <<"0 G pmlinewidth "<<SizeY<<" div w 0 0 M 1 0 V 0 1 V -1 0 V C stroke\n";
double tmp=MIN(GRSizeX,GRSizeY)*3200/FontScale;
*out <<"0.1 0 M 0 G /Helvetica findfont "
     <<iround(tmp)
     <<" ScaleFont setfont\ngsave ";
//out->precision(6); *out << 1/SizeX<<' '<<1/SizeY<<" scale 0 " <<-iround(tmp*1.5)<<' '<<ScaleFont<<" mul R (";
out->precision(6); *out << 1/SizeX<<' '<<1/SizeY<<" scale 0 "<<-iround(tmp*0.5)<<" R (";
double z = isZlinlog ? alog10(Zfrom):Zfrom;
sprintf(charTmp,"%.4g",z); *out<<charTmp;
/* // old version (before 4.4.1998)
if (z>1e4 || z<1e-2) {
    out->setf(ios::scientific, ios::floatfield); out->precision(3); }
  else {
    out->setf(ios::fixed, ios::floatfield);
    if (z>1e2) out->precision(1); else
    if (z>1) out->precision(2); else out->precision(4);
    }
*out << z;
*/
out->setf(ios::fixed, ios::floatfield); out->precision(3);
*out <<") showTopAlign grestore\n0.1 1 M ";
out->precision(6); *out << 1/SizeX<<' '<<1/SizeY<<" scale 0 "<< tmp*iround(0.5) << " R (";
z = isZlinlog ? alog10(Zto):Zto;
sprintf(charTmp,"%.4g",z); *out<<charTmp;
/* // old version (before 4.4.1998)
out->setf(ios::scientific, ios::floatfield); out->precision(3);
if (z>1e6 || z<1e-2) {
    out->setf(ios::scientific, ios::floatfield); out->precision(3); }
  else {
    out->setf(ios::fixed, ios::floatfield);
    if (z>1e4) out->precision(0); else
    if (z>1e2) out->precision(1); else
    if (z>1) out->precision(2); else out->precision(4);
    }
*out << z;
*/
*out<<") show\ngrestore\n";
out->setf(ios::fixed, ios::floatfield); out->precision(6);
if (!*out) {
  cerr<<"Problem writing to output file.\n";
  if (++outproblems>=MaxOutProblems) {
    cerr <<"Too much output problems (disk full? disk quota exceeded? pipe closed?) Terminating.\n";
    return; }
  }
SetColor(-1);
}


void DoKeyOfDraw3D ()
{
// make intensity boxes with labels
if (ZkeyIntervals==0) return;
if (ZkeyIntervals<0) { DoKeyOfDraw3D_smooth_box(); return; }
if (ZkeyIntervals==1) ZkeyIntervals=2;
double Z=Zto, zstep=(Zto-Zfrom)/(ZkeyIntervals-1);
double RefX=RelRefX, RefY=RelRefY; // reference point wrt graph
double dx=0.075, dy=0.02;          // box size wrt graph size
long xA,xB,yA,yB;
*out <<"\t\t\t\t\t\t%now gray boxes with keys\n";
while (Z>Zfrom-zstep*0.5) {
  if (Z<Zfrom+zstep*0.5) Z=Zfrom; // avoid rounding problems
  out->precision(0);
  xA=long( wrtGraphX(RefX) );    yA=long( wrtGraphY(RefY) );
  xB=long( wrtGraphX(RefX+dx) ); yB=long( wrtGraphY(RefY-dy) );
  SetColor(Get3dColor(Z));
  *out << xA <<' '<< yA <<" N " // fill box
       << xB-xA <<" 0 V ";
  *out << " 0 "<< yB-yA <<" V "
      << xA-xB <<" 0 s\n";
  RefY-=3.*dy;  // corresponding number
  *out<<"0 G "; font(2800);
  *out<<long( wrtGraphX(RefX+dx*0.5) )<<' '<<long( wrtGraphY(RefY) )<<" M (";
  out->precision(4); *out << (isZlinlog ? alog10(Z):Z) << ") Cshow\n";
  if (!*out) {
    cerr<<"Problem writing to output file.\n";
    if (++outproblems>=MaxOutProblems) {
      cerr <<"Too much output problems (disk full? disk quota exceeded? pipe closed?) Terminating.\n";
      return; }
    }
  RefY-=2.*dy;
  Z-=zstep;
  }
}


void TerminateOut ()
{
// the last things before closing the output file
long xA,yA,xB,yB;
#ifdef RelativeRectangles
*out << "M\n";
#endif
*out << "%pm3d_map_end\n";
*out << "grestore\t% terminate clipping\n\n"; // TerminateClip;
MakeGridXY();
*out << "TicsX\n";
*out << "TicsY\n";
DoKeyOfDraw3D();  // moved here: 13.11.1996
if (isboundary) { // draw boundary on and around the graph
  *out <<"\n0 G pmlinewidth w\t\t\t%now boundary around the graph\n";
  xA=long( floor(x2p(Xfrom)) ); yA=long( floor(y2p(Yfrom)) );
  xB=long( floor(x2p(Xto)) );   yB=long( floor(y2p(Yto)) );
  *out << xA <<' '<< yA <<" N ";
  *out << xB-xA <<' '<< 0 <<" V ";
  *out << 0 <<' '<< yB-yA <<" V ";
  *out << xA-xB <<' '<< 0 <<" V C stroke\n";
  }
ifFF {
  cerr <<"Data file has "<<Inp.line<<" lines and "<<Scans<<" scans, "
	  "max. nb of pts in one scan "<<MaxAimax<<'\n' CERR_FLUSH;
  Inp.close(); }
WriteTail();
if (CanCloseOut) {
  if (*OutFile=='|')
      { out->flush(); cout << "Whole picture drawn. "; PressEnter(); }
    else
      WriteTail2();  // write "showpage" -- clear the page
  ((ofstream*)out)->close();
  }
// *out << char('Z'-64); //?
if (WaitKillOut) {
  // press Ctrl/C to kill it
  cerr << "Whole picture drawn. Press <return or Ctrl/C> to continue...\n";
  cin.ignore(32535,'\n');
  }
}


void NotMemoryClose ()
{
cerr <<"Not enough memory, sorry.\n";
TerminateOut();
exit(1);
}


void EnlargeArrays ()
{
// it enlarges all the arrays storing the scans by ArrayDelta
ArrayMax += ArrayDelta;
/*  // my debugging message:
cerr <<"Technical note: enlarging scan arrays to accept up to "
     <<ArrayMax<<" pts at line "<<Inp.line<<endl CERR_FLUSH; */
Ax = (double*) realloc(Ax, sizeof(double)*ArrayMax);
Ay = (double*) realloc(Ay, sizeof(double)*ArrayMax);
Az = (double*) realloc(Az, sizeof(double)*ArrayMax);
Bx = (double*) realloc(Bx, sizeof(double)*ArrayMax);
By = (double*) realloc(By, sizeof(double)*ArrayMax);
Bz = (double*) realloc(Bz, sizeof(double)*ArrayMax);
if (!Ax || !Ay || !Az || !Bx || !By || !Bz) NotMemoryClose();
Ain = (char*) realloc(Ain, sizeof(char)*ArrayMax);
Bin = (char*) realloc(Bin, sizeof(char)*ArrayMax);
if (!Ain || !Bin) NotMemoryClose();
}


void FillBin ( word bimax )
{
// It fills the Bin array
for (word i=bimax; i--; )
  Bin[i] = (Bx[i]>=Xfrom && Bx[i]<=Xto) &&
	   (By[i]>=Yfrom && By[i]<=Yto);
}


char ReadScan ( word& bimax )
{
// read the next scan to Bx,By,Bz arrays.
// Number of points is bimax
// it returns 0 if all OK, 2 if eof
// Technical note: arrays Ax..Az or Bx..Bz cannot be parameters to this
//   routine. That's why they are global
double X,Y,Z;
char result = 0;
Scans++;
bimax = 0; ThisEveryNth = 1;
while (Inp.eoln_skipspaces())
  Inp.xligno();
while (isFFand !Inp.eoln_skipspaces())
// || (isFromWhich==isFromFormula && (FromFormula^.IY<FromFormula^.MaxY) && (FromFormula^.IX==0))
  {
  if (isFFand Inp.eof()) { result = 2; break; }
  result = ReadTheLine(X,Y,Z,Inp,3);
  if (result) break; // eof or new scan
  if (isZlinlog) Z = Z>0 ? log10(Z) : -38;
  if (bimax>=ArrayMax) EnlargeArrays();
  Bx[bimax] = X;
  By[bimax] = Y;
  Bz[bimax] = Z;
  bimax++;
  }
if (SmoothEveryNth > 1) {
    int i, r, chunk;
    for (i=r=0; r<bimax; i++) {
      Bx[i] = Bx[r];
      By[i] = By[r];
      Bz[i] = Bz[r++];
      for (chunk=1; chunk<SmoothEveryNth && r<bimax; chunk++, r++) {
	Bx[i] += Bx[r];
	By[i] += By[r];
	Bz[i] += Bz[r]; }
      if (chunk > 1) {
	Bx[i] /= chunk;
	By[i] /= chunk;
	Bz[i] /= chunk; }
      }
    bimax = i;
#if 0
    ofstream O; O.open("smooth.log",ios::app);
    O <<"# bimax=" <<bimax<<'\n';
    for (i=0; i<bimax; i++) O <<Bx[i] <<'\t' <<By[i] <<'\t' <<alog10(Bz[i]) <<'\n';
    O <<'\n';
    O.close();
#endif
  }
if (result==1) return 0;
return result;
}


// ** several small routines for command line params input management:

char PaperRequired[] = "paper size: DOS or DOSL or A4 or A4L";
char PaperNameCalled = 0;

char PaperName2Values ( char * charTmp )
{
// Convert paper name into numerical values of the page size
// and page shift found by the trial and error method
// return 0 if all OK, 1 if error
char NoOK=0;
landscape=0;
if (!strcmp(strlwr(charTmp),"dos"))  { // ghostscript under dos
    #ifdef gs261
    // old values. It worked with gs 2.61, now it has changed?
    PaperX=179200; PaperY=134314;
    ShiftX0=0.0010; ShiftY0=0.0014;
    FontScale=0.7;
    #else
    // now I use gs 3.33
    PaperX=110000; PaperY=82000;
    ShiftX0=0.0025; ShiftY0=0.035;
    FontScale=1.25;
    #endif
    landscape=0; } else
if (!strcmp(strlwr(charTmp),"dosl")) { // ghostscript under dos, landscape
    #ifdef gs261
    PaperX=134314; PaperY=179200;
    ShiftX0=0.0014; ShiftY0=0.0010;
    FontScale=0.7;
    #else
    PaperX=84000; PaperY=105000;
    ShiftX0=-0.02; ShiftY0=0.05;
    FontScale=1.25;
    #endif
    landscape=1; } else
if (!strcmp(strlwr(charTmp),"a4"))   { // a4 paper
    // PaperX=179200*1/1.1351*0.287*1.3125*1.08/1.0774;  PaperY=134314*1/1.2712*0.4541*1.7763*1.144/1.02/1.025/1.5/0.744/1.015;
    PaperX=59611.75; PaperY=82327.89/0.97186;
    // when debug: multiplying by x>1 the picture incr.
    ShiftX0=0.0019; ShiftY0=0.00647;
    FontScale=2.2;
    landscape=0; } else
if (!strcmp(strlwr(charTmp),"a4l"))  { // a4 paper, landscape
    PaperX=82327.89/0.97186; PaperY=59611.75;
    ShiftX0=0.00647; ShiftY0=0.0019;
    FontScale=2.2;
    landscape=1; }
else NoOK=1;
PaperNameCalled=1;
return NoOK;
}

void opt1double( double& X )
{
// this read X from the command line (from optind)
char *c;
X = strtod(optarg,&c);
if (c[0]) { cerr <<"Wrong number " <<optarg <<endl; exit(1); }
}


void opt2double( double& From, double& To )
{
// this read From, To from the command line (from optind)
char *c;
From = strtod(optarg,&c);
if (c[0]) { cerr <<"Wrong number " <<optarg <<endl; exit(1); }
To = strtod(optarg2,&c);
if (c[0]) { cerr <<"Wrong number " <<optarg2 <<endl; exit(1); }
}

void opt3read (long& FromLine, long& FromScan, int& EveryNth)
{
char *c;
FromLine = strtol(optarg,&c,10);
if (c[0]) { cerr<<"Wrong number "<<optarg<<endl; exit(1); }
FromScan = strtol(optarg2,&c,10);
if (c[0]) { cerr <<"Wrong number " <<optarg2 <<endl; exit(1); }
EveryNth = strtol(optarg3,&c,10);
if (c[0]) { cerr <<"Wrong number " <<optarg3 <<endl; exit(1); }
}

void opt4double (double& a, double& b, double& c)
{
char *cc;
a=strtod(optarg2,&cc);
if (cc[0]) { cerr <<"Wrong number " <<optarg2 <<endl; exit(1); }
b=strtod(optarg3,&cc);
if (cc[0]) { cerr <<"Wrong number " <<optarg3 <<endl; exit(1); }
c=strtod(optarg4,&cc);
if (cc[0]) { cerr <<"Wrong number " <<ansicpp_argv[optind] <<endl; exit(1); }
}


void optNotes ()
{
readnotes = tolower(optarg[0]) == 'y';
if (readnotes) {
  if (optind==ansicpp_argc) { cerr <<"Missing x or y -N Y\n"; exit(1); }
  optarg=ansicpp_argv[optind++];
  optarg2=ansicpp_argv[optind++];
  opt2double(NotesX,NotesY);
/*  char *c;
  NotesX=strtod(optarg2,&c);
  if (c[0]) { cerr<<"Wrong number "<<optarg2<<endl; exit(1); }
  NotesY=strtod(optarg3,&c);
  if (c[0]) { cerr<<"Wrong number "<<optarg3<<endl; exit(1); }
*/
  }
}


void opt2grayboxes ()
{
char *c;
ZkeyIntervals = strtol(optarg,&c,10);
if (c[0]) { cerr <<"Wrong number "<<optarg <<endl; exit(1); }
RelRefX = strtod(optarg2,&c);
if (c[0]) { cerr <<"Wrong number "<<optarg2 <<endl; exit(1); }
RelRefY = strtod(optarg3,&c);
if (c[0]) { cerr <<"Wrong number "<<optarg3 <<endl; exit(1); }
}


void ColorFormulaParamError ( char *S, int Line )
{
cerr <<"Wrong parameter for colouring: " <<S;
if (Line>=0) cerr <<" on line " <<Line;
cerr<<endl;
exit(1);
}


void ColorFormulaOutOfRange ( char *S, int& formulaRGB )
{
cerr <<"Warning: parameter for colouring out of range " <<S <<": "
     <<formulaRGB <<">=" <<colorFormulae <<"; reset to 0\a\n";
formulaRGB = 0;
}


void processColorFormulaParameter ( char *S, int Line=-1 )
{
char *S1=S, *c; long L1,L2,L3;
L1=strtol(S1,&c,10);
if (*c!=';') ColorFormulaParamError(S,Line);
S1=c+1; L2=strtol(S1,&c,10);
if (*c!=';') ColorFormulaParamError(S,Line);
S1=c+1; L3=strtol(S1,&c,10);
if (*c) ColorFormulaParamError(S,Line);
formulaR=L1; formulaG=L2; formulaB=L3;
if (abs(formulaR)>=colorFormulae) ColorFormulaOutOfRange(S,formulaR);
if (abs(formulaG)>=colorFormulae) ColorFormulaOutOfRange(S,formulaG);
if (abs(formulaB)>=colorFormulae) ColorFormulaOutOfRange(S,formulaB);
}


void SetEveryNth ()
{
// if EveryNth is negative, then set it to 1 and use the previous value
// for smoothing every Nth
if (EveryNth == 0 || EveryNth == -1) {
  EveryNth = 1; return; }
if (EveryNth > 0) return;
// now it is negative
SmoothEveryNth = -EveryNth;
EveryNth = 1;
}


void  AutoScale ( char x, char y, char z )
{
// makes autoscale. The input files must be already defined
double XX, YY, ZZ; int i;
double MinX=1e308,Min2X=1e308,MinY=1e308,Min2Y=1e308,MinZ=1e308,Min2Z=1e308;
double MaxX=-1e308,Max2X=-1e308,MaxY=-1e308,Max2Y=-1e308,MaxZ=-1e308,Max2Z=-1e308;
  // note:some functions in pmparser return 1e-308 if sth wrong in the expression

cerr <<"Autoscale: reading from " <<InpFile <<endl;
Inp.open(InpFile);
#ifndef __TURBOC__ // some problems
#if !(defined(__GNUC__) && __GNUC__>=3)
Inp.setbuf(input_buffer,size_input_buffer);
#endif
#endif
for (i=FromLine; --i>0; Inp.xligno());
ThisEveryNth = 1;

while (!Inp.eof_skipspaces()) {
  ReadTheLine(XX,YY,ZZ,Inp,1);
  if (Inp.fail()) break;
  if (x) {
    if (XX<MinX) { Min2X=MinX; MinX=XX; } else
      if (XX<Min2X) { if (XX>MinX) Min2X=XX; }
    if (XX>MaxX) { Max2X=MaxX; MaxX=XX; } else
      if (XX>Max2X && XX<MaxX) Max2X=XX;
    }

  if (y) {
    if (YY<MinY) { Min2Y=MinY; MinY=YY; } else
      if (YY<Min2Y) { if (YY>MinY) Min2Y=YY; }
    if (YY>MaxY) { Max2Y=MaxY; MaxY=YY; } else
      if (YY>Max2Y && YY<MaxY) Max2Y=YY;
    }

  if (z) {
    if (ZZ<MinZ) { Min2Z=MinZ; MinZ=ZZ; } else
      if (ZZ<Min2Z) { if (ZZ>MinZ) Min2Z=ZZ; }
    if (ZZ>MaxZ) { Max2Z=MaxZ; MaxZ=ZZ; } else
      if (ZZ>Max2Z && ZZ<MaxZ) Max2Z=ZZ;
    }
  } // while not eof

Inp.close();  // end of reading
if (x) { if (Max2X<Min2X) Max2X=Min2X;
	 if (MaxX<Max2X) MaxX=Max2X; }
if (y) { if (Max2Y<Min2Y) Max2Y=Min2Y;
	 if (MaxY<Max2Y) MaxY=Max2Y; }
if (z) { if (Max2Z<Min2Z) Max2Z=Min2Z;
	 if (MaxZ<Max2Z) MaxZ=Max2Z; }

cerr <<"** Results of autoscaling (min and max): **\n";
if (x) {
  if (ColumnX>0) cerr <<"Column " <<ColumnX; else cerr <<FormulaX;
  cerr <<":\t" <<MinX <<"\t" <<Min2X <<" ..\t" <<Max2X <<"\t" <<MaxX <<"\n";
  Xfrom=MinX; Xto=MaxX;
  if (Xfrom<-0.9e308) Xfrom=Min2X; // wrong margins
  if (Xto>0.9e308) Xto=Max2X;
  if (Xfrom>=Xto) Xfrom=-2e308;
  }
if (y) {
  if (ColumnY>0) cerr <<"Column " <<ColumnY; else cerr <<FormulaY;
  cerr <<":\t" <<MinY <<"\t" <<Min2Y <<" ..\t" <<Max2Y <<"\t" <<MaxY <<"\n";
  Yfrom=MinY; Yto=MaxY;
  if (Yfrom<-0.9e308) Yfrom=Min2Y; // wrong margins
  if (Yto>0.9e308) Yto=Max2Y;
  if (Yfrom>=Yto) Yfrom=-2e308;
  }
if (z) {
  if (ColumnZ>0) cerr <<"Column " <<ColumnZ; else cerr <<FormulaZ;
  cerr <<":\t" <<MinZ <<"\t" <<Min2Z <<" ..\t" <<Max2Z <<"\t" <<MaxZ <<"\n";
  Zfrom=MinZ; Zto=MaxZ;
  if (Zfrom<-0.9e308) Zfrom=Min2Z; // wrong margins
  if (Zto>0.9e308) Zto=Max2Z;
  if (Zfrom>=Zto) Zfrom=-2e308;
  }
if (isZlinlog) {
  if (Zfrom<=0) { Zfrom=Min2Z; if (Zfrom<=0) Zfrom=1e-35; }
  if (Zto<=0) Zto=1e35; if (Zfrom>Zto) Zto*=10;
  }
} // of AutoScale



void Help ( int exit_code );


// *** MAIN PART: GrafFrom3D ***

void GrafFrom3D ( int argc, char *argv[] )
{
char charTmp[256];
long LL;
double GX1=0,GY1=0;
long iGX1=0,iGY1=0, iGX2,iGY2, iGX3,iGY3, iGX4,iGY4;
#ifdef RelativeRectangles
long iGXfix,iGYfix; char firstsprintf=1;
#endif
char Rods=0;
int iWidth,iWidth2;
double RodsWidth;
char Speed=1;
int Argc=1; char **Argv=0; int save_optind;

// ** read file  pm3d.p3d
optind=1;
if (!strchr(PMGETOPT_SW,argv[optind][0])) {  // SW="-/" from getopt.cpp
// read the input file.p3d:
Inp.open(argv[1]); // 1. 11. 1998: open in binary mode
if (!Inp)
  { cerr <<"Cannot open file "<<argv[1]<<" !\n"; exit(1); }
cerr << "Reading setup file "<<argv[1];

Inp >> charTmp;
if (!strcmp(charTmp,"pm3d_version:")) goto NewJpm3dVersion;
// older version --- the pre-Java version
cerr << " (older version detected)\n" CERR_FLUSH;
Inp.ligno();
 InpFile = strdup(charTmp);
Inp >> charTmp; Inp.ligno();
 OutFile = strdup(charTmp);
Inp >> FromLine; Inp.ligno();
Inp >> FromScan; Inp.ligno();
Inp >> EveryNth; SetEveryNth(); Inp.ligno();
ReadColumnsFromInp (Inp,3); Inp.ligno();
Inp >> Xfrom >> Xto; Inp.ligno();
Inp >> Yfrom >> Yto; Inp.ligno();
Inp >> Zfrom >> Zto; Inp.ligno();
Inp >> isZlinlog;    Inp.ligno();
// now "fligna Amalka": grid is default and can be omitted (old .p3d files)
Inp >> charTmp;
if (!strcmp(strlwr(charTmp),"grid")) { //this is default
  Inp.ligno(); Inp >> charTmp; }
else
if (!strcmp(strlwr(charTmp),"tr")) { //truncation rods
  Rods=1; Inp >> RodsWidth;
  if (RodsWidth==0) Inp.required("Non-zero Rod Width");
  Inp.ligno(); Inp >> charTmp; }
if (PaperName2Values(charTmp))
  Inp.required(PaperRequired); Inp.ligno();
Inp >> GRSizeX >> GRSizeY; Inp.ligno(); // graph relative size
Inp >> GX1 >> GY1; Inp.ligno(); // additional relative shifts
Inp >> charTmp; FormulaParserPreProcess(charTmp);
  if (!strcmp(charTmp,"x")) GrayFormula = 0;
    else GrayFormula = strdup(charTmp);
  Inp.ligno();
Inp >> isnegative;
  switch (isnegative) {
    case 'p': case 'P': isnegative=0; break;
    case 'n': case 'N': isnegative=1; break;
    default: Inp.required("Positive or Negative"); }
  Inp.ligno();
Inp >> isbackblack;
  switch (isbackblack) {
    case 'w': case 'W': isbackblack=0; break;
    case 'b': case 'B': isbackblack=1; break;
    default: Inp.required("White or Black background"); }
  Inp.ligno();
Inp >> isboundary; Inp.ligno();
  Inp.ligno(); // skip the obligatory remark
Inp.get(Title,sizeof(Title));   Inp.ligno();
Inp.get(LabelX,sizeof(LabelX)); Inp.ligno();
Inp.get(LabelY,sizeof(LabelY)); Inp.ligno();
Inp >> readnotes; readnotes=tolower(readnotes)=='y';
  if (readnotes) { Inp >> NotesX >> NotesY; }
  Inp.ligno();
Inp >> ZkeyIntervals >> RelRefX >> RelRefY;
  if (ZkeyIntervals<0) {
    if (GrayFormula) {
      cout<<"Cannot use smooth boxes together with gray formula, sorry.\n";
      exit(1); }
    Inp >> RelBoxSizeX >> RelBoxSizeY; }
Inp.ligno();

// now syntax: either "end", or additional "command-line options"
// Inp.testEnd();
Inp >> charTmp;
if (stricmp(charTmp,"END")) {
  // well, additional command line opts are given
  iWidth = Inp.line;
  Argv = new char* [50];
  Argv[0] = strdup(argv[1]); // and let's ignore string limitations...
  // Argc=1; set up above
  while (1) {
    if (!stricmp(charTmp,"END")) break;
    if (iWidth<(int)Inp.line-1)
      { cout << "END required on line "<<iWidth+1<<" !\n"; exit(1); };
    Argv[Argc] = strdup(charTmp); // and let's ignore string limitations...
    Argc++;
    Inp >> charTmp;
    }
  }
goto AfterNewJpm3dVersion;
// end of reading the older version


NewJpm3dVersion: ;
// ***************** ReadFile_Jp3d (i.e., the new format) ********************
//char flag=0;
int i; char S[255];
cerr << " (new version detected)\n" CERR_FLUSH;
Inp >> charTmp; // skip the version

while (1) {
  Inp >> S;
  // if (flag<0) break;
  // System.err.println("Dalsi slovo: "+S);
  if (!strcmp(S,"input_data_file:")) {
	Inp.readStringWithSpaces( charTmp, sizeof(charTmp) );
	// Inp.ligno();
	InpFile = strdup(charTmp);
	continue; }
  if (!strcmp(S,"output_ps_file:")) {
	Inp.readStringWithSpaces( charTmp, sizeof(charTmp) );
	// Inp.ligno();
	OutFile = strdup(charTmp);
	continue; }
  if (!strcmp(S,"p3d_file:")) {
	Inp >> charTmp; /*read but ignore*/
	continue; }
  if (!strcmp(S,"from_line:")) {
	Inp >> FromLine; continue; }
  if (!strcmp(S,"from_scan:")) {
	Inp >> FromScan; continue; }
  if (!strcmp(S,"take_n-th_point:")) {
	Inp >> EveryNth; SetEveryNth(); continue; }
  if (!strcmp(S,"plot_2line_note:")) {
	Inp >> S; S[0]=tolower(S[0]);
	readnotes = !S[1] && (tolower(S[0])=='y' || S[0]=='1');
	continue; }
  if (!strcmp(S,"plot_2line_note_at:")) {
	Inp >> NotesX >> NotesY; continue; }
  if (!strcmp(S,"columns:")) {
	ReadColumnsFromInp (Inp,3); continue; }
  if (!strcmp(S,"rangeX:")) {
	Inp >> Xfrom >> Xto; continue; }
  if (!strcmp(S,"rangeY:")) {
	Inp >> Yfrom >> Yto; continue; }
  if (!strcmp(S,"rangeZ:")) {
	Inp >> Zfrom >> Zto; continue; }
  if (!strcmp(S,"logZ:")) {
	Inp >> S; strlwr(S);
	isZlinlog = !S[1] && (tolower(S[0])=='y' || S[0]=='1');
	continue; }
  if (!strcmp(S,"title:")) {
	Inp.get(Title,sizeof(Title)); Inp.ligno();
	continue; }
  if (!strcmp(S,"label_x:")) {
	Inp.get(LabelX,sizeof(LabelX)); Inp.ligno();
	continue; }
  if (!strcmp(S,"label_y:")) {
	Inp.get(LabelY,sizeof(LabelY)); Inp.ligno();
	continue; }
  if (!strcmp(S,"paper_format:")) {
	Inp >> charTmp;
	if (PaperName2Values(charTmp));
	continue; }
  if (!strcmp(S,"graph_rel_size:")) {
	Inp >> GRSizeX >> GRSizeY; continue; }
  if (!strcmp(S,"graph_rel_shift:")) {
	Inp >> GX1 >> GY1; continue; }
  if (!strcmp(S,"color_map:")) {
	Inp >> ColorMode;
	if (!strchr(availableCOLORMODES,ColorMode))
	  Inp.required("colour modes: g (gray), r (rgb)");
	continue; }
  if (!strcmp(S,"color_rgb_param:")) {
    Inp >> S; processColorFormulaParameter(S,Inp.line);
    continue; }
  if (!strcmp(S,"gray_formula:")) {
	Inp >> charTmp; FormulaParserPreProcess(charTmp);
	  if (!strcmp(charTmp,"x")) GrayFormula=0;
	    else GrayFormula = strdup(charTmp); 
	continue; }
  if (!strcmp(S,"picture_photo:")) {
	Inp >> charTmp; strlwr(charTmp);
	isnegative = !strcmp(charTmp,"negative");
	continue; }
  if (!strcmp(S,"background:")) {
	Inp >> charTmp; strlwr(charTmp);
	isbackblack = !strcmp(charTmp,"black");
	continue; }
  if (!strcmp(S,"draw_graph_boundary:")) {
	Inp >> S;
	isboundary = !S[1] && (tolower(S[0])=='y' || S[0]=='1');
	continue; }
  if (!strcmp(S,"gray_box:")) {
	Inp >> S; strupr(S);
	if (!strcmp(S,"NO")) ZkeyIntervals=0; else
	if (!strcmp(S,"DISCRETE")) ZkeyIntervals = (ZkeyIntervals==0) ? 8:abs(ZkeyIntervals); else
	if (!strcmp(S,"SMOOTH")) ZkeyIntervals = (ZkeyIntervals==0) ? -8:-abs(ZkeyIntervals);
	continue; }
  if (!strcmp(S,"gray_box_levels:")) {
	Inp >> i;
	if (ZkeyIntervals>0) ZkeyIntervals=i; else
	if (ZkeyIntervals<0) ZkeyIntervals=-i;
	continue; }
  if (!strcmp(S,"gray_box_pos:")) {
	Inp >> RelRefX >> RelRefY; continue; }
  if (!strcmp(S,"gray_box_size:")) {
	Inp >> RelBoxSizeX >> RelBoxSizeY; continue; }
  if (!strcmp(S,"add_cmd_opts:")) {
	cerr << "WARNING: add_cmd_opts are not supported in new format.\n";
	Inp.ligno();
	continue; }
  if (!strcmp(S,"flush_scans:")) {
	Inp >> S; Flush=toupper(S[0]);
	if (!strchr("LRC",Flush)) Flush='L';
	continue; }
  if (!strcmp(S,"try_reduce_size:")) {
	Inp >> S;
	Speed = !( !S[1] && (tolower(S[0])=='y' || S[0]=='1') );
	continue; }
  if (!strcmp(S,"rel_font_size:")) {
	Inp >> ScaleFont;
	if (ScaleFont<=0) ScaleFont=1; continue; }
  if (!strcmp(S,"autoscale:")) {
	Inp >> S; strlwr(S);
	DoAutoscaleX=strchr(S,'x')!=NULL;
	DoAutoscaleY=strchr(S,'y')!=NULL;
	DoAutoscaleZ=strchr(S,'z')!=NULL;
	if (strlen(S)>3 || !strchr("xyz",S[0]) ||
		  (S[1] && (!strchr("xyz",S[1]) ||
		    (S[2] && !strchr("xyz",S[2]))) )) {
		  cerr<<"Autoscale option requires  xyz  combination\n";
		  exit(1); }
	continue; }
  if (!strcmp(S,"algorithm:")) {
	Inp >> S; Rods = !strcmp(strupr(S),"TR"); continue; }
  if (!strcmp(S,"algorithm_TR_width:")) {
	Inp >> RodsWidth; if (RodsWidth==0) RodsWidth=0.5; continue; }
  if (!strcmp(S,"algorithm_TR_width:")) {
	Inp >> RodsWidth; continue; }
  if (!strcmp(S,"user_ticks_x:")) {
	Inp >> S; strlwr(S);
	XticksOn = !S[1] && (tolower(S[0])=='y' || S[0]=='1');
	continue; }
  if (!strcmp(S,"user_ticks_y:")) {
	Inp >> S; strlwr(S);
	YticksOn = !S[1] && (tolower(S[0])=='y' || S[0]=='1');
	continue; }
  if (!strcmp(S,"ticks_x_range:")) {
	Inp >> XticksFrom >> XticksStep >> XticksTo;
	continue; }
  if (!strcmp(S,"ticks_y_range:")) {
	Inp >> YticksFrom >> YticksStep >> YticksTo;
	continue; }
  if (!strcmp(S,"ticks_out:")) {
	Inp >> S;
	TicksOut = !S[1] && (tolower(S[0])=='y' || S[0]=='1');
	continue; }
  if (!strcmp(S,"vertical_yticks_labels:")) {
	Inp >> S;
	VerticalYTicksLabels = !S[1] && (tolower(S[0])=='y' || S[0]=='1');
	continue; }
  if (!strcmp(S,"minor_x_freq:")) {
	Inp >> MinorTicksX; continue; }
  if (!strcmp(S,"minor_y_freq:")) {
	Inp >> MinorTicksY; continue; }
  if (!strcmp(S,"where_minorX:")) {
	Inp >> WhereMinorX;
	if (WhereMinorX<0 || WhereMinorX>2) WhereMinorX=0;
	continue; }
  if (!strcmp(S,"where_minorY:")) {
	Inp >> WhereMinorY;
	if (WhereMinorY<0 || WhereMinorY>2) WhereMinorY=0;
	continue; }
  if (!strcmp(S,"major_gridX:")) {
	Inp >> S;
	MajorGridX = !S[1] && (tolower(S[0])=='y' || S[0]=='1');
	continue; }
  if (!strcmp(S,"major_gridY:")) {
	Inp >> S;
	MajorGridY = !S[1] && (tolower(S[0])=='y' || S[0]=='1');
	continue; }
  if (!strcmp(S,"when_gridX:")) { Inp >> charTmp;
    switch (tolower(charTmp[0])) {
      case 'n': whenGridX=gridNever; break;
      case 'b': whenGridX=gridBefore; break;
      case 'a': whenGridX=gridAfter; break;
      default: charTmp[1]=' '; }
    if (strlen(charTmp)!=1) {
      cerr<<"options for grid: n, b, a  (never, before, after)\n";
      exit(1); }
    continue; }
  if (!strcmp(S,"when_gridY:")) { Inp >> charTmp;
    switch (tolower(charTmp[0])) {
      case 'n': whenGridY=gridNever; break;
      case 'b': whenGridY=gridBefore; break;
      case 'a': whenGridY=gridAfter; break;
      default: charTmp[1]=' '; }
    if (strlen(charTmp)!=1) {
      cerr<<"options for grid: n, b, a  (never, before, after)\n";
      exit(1); }
    continue; }
  // The End:
  if (!strcmp(strupr(S),"END")) break;
  if (strlen(S)>0) cerr << "Unknown label: "<< S << endl CERR_FLUSH;
  }
// The End of Reading New Jp3d File
AfterNewJpm3dVersion: ;

Inp.close();
optind++;
}


/********   now read the command line options   ****/
signed char c;
char DoBlankLine=0;
// now the stuff for the complicated line concatenation
char isTail=0;
char Line[255], OldLine[255]="", Tail[255];
int LineLen, OldLineLen, TailLen;
char *LineEnd, *OldLineEnd, *cptr;
double LineColor, OldLineColor;
// stuff for rods
int OldiWidth=-12345;
char CanCopyScan=0;
double tmp, theshift;

char flag_opt=0;
if (Argc==1) { Argc=argc; Argv=argv; flag_opt=1; }
  else { save_optind=optind; optind=1; ansicpp_argc=Argc; ansicpp_argv=Argv; }

while (Argc>1) {
while ((c = pmgetopt (Argc, Argv, "i:o:x;y;z;X:Y:Z:G:c:C:g;s;r!L:T:N:a:P:R:K!F:f:t_pNlBbnwhS?")) != EOF) {
  switch (c) {
    case 'P': if (PaperName2Values(optarg))
      { cerr <<"Required: "<<PaperRequired<<endl; exit(1); } break;
    case 'n': isnegative=isnegative==0; break;
    case 'l': isZlinlog=isZlinlog==0; break;
    case 'i': if (InpFile) free(InpFile);
	      InpFile = strdup(optarg); break;
    case 'o': if (OutFile) free(OutFile);
	      OutFile = strdup(optarg); break;
    case 'x': opt2double(Xfrom,Xto); break;
    case 'y': opt2double(Yfrom,Yto); break;
    case 'z': opt2double(Zfrom,Zto); break;
    case 'g': opt2double(GRSizeX,GRSizeY); break;
    case 's': opt2double(GX1,GY1); break;
    case 'c': if (!strchr(availableCOLORMODES,*optarg)) {
		cerr <<"Available colour modes: " <<colorModeGRAY <<" (gray), "
		     <<colorModeRGB <<" (rgb)\n";
		exit(1); }
	      ColorMode = *optarg; break;
    case 'C': processColorFormulaParameter(optarg); break;
    case 'G': strcpy(charTmp,optarg); FormulaParserPreProcess(charTmp);
	      if (GrayFormula!=0) free(GrayFormula);
	      if (!strcmp(charTmp,"x")) GrayFormula = 0;
		else GrayFormula = strdup(charTmp);
    case 'X': String2Column(optarg,1); break;
    case 'Y': String2Column(optarg,2); break;
    case 'Z': String2Column(optarg,3); break;
    case 'r': opt3read(FromLine,FromScan,EveryNth); SetEveryNth(); break;
    case 'B': isbackblack = isbackblack==0; break;
    case 'b': isboundary = isboundary==0; break;
    case 'L': strcpy(LabelX,optarg);
	      if (optind==ansicpp_argc) { cerr <<"Missing label!\n"; exit(1); }
	      strcpy(LabelY,argv[optind++]); break;
    case 'T': strcpy(Title,optarg); break;
    case 'N': optNotes(); break;
    case 'a': if (strlen(optarg)>3 || !strchr("xyz",optarg[0]) ||
		  (optarg[1] && (!strchr("xyz",optarg[1]) ||
		    (optarg[2] && !strchr("xyz",optarg[2]))) )) {
		  cerr<<"Autoscale option requires  xyz  combination\n";
		  exit(1); }
		strlwr(optarg);
		DoAutoscaleX = strchr(optarg,'x')!=NULL;
		DoAutoscaleY = strchr(optarg,'y')!=NULL;
		DoAutoscaleZ = strchr(optarg,'z')!=NULL;
		break;
    case 'K': opt2grayboxes(); break;
    case 'w': WaitKillOut=1; break;
    case 'S': Speed=0; break;
    case 'R': Rods=1; opt1double(RodsWidth);
	      if (RodsWidth==0) { cerr <<"Rod Width cannot be 0 !\n"; exit(1); }
	      break;
    case 'F': if (strlen(optarg)>1 || !strchr("LRC",optarg[0]))
		{ cerr <<"Flush scans: Left, Right or Center\n"; exit(1); }
	      Flush=optarg[0]; break;
    case 'f': opt1double(ScaleFont);
	      if (ScaleFont<=0) { cerr <<"Font scale must be positive!\n"; exit(1); }
	      break;
    case 't': switch (toupper(*optarg)) {
		case 'X': opt4double(XticksFrom,XticksStep,XticksTo);
			  XticksOn=1; break;
		case 'Y': opt4double(YticksFrom,YticksStep,YticksTo);
			  YticksOn=1; break;
		default:  cout <<"parameters of ticks: x from step to | y from to step\n";
			  exit(1);
		}
	      break;
    case 'h': case '?': Help(0); // error in the command line option
    case PMGETOPT_BADCH: exit(1);
    }
  } // of while c=getopt
if (flag_opt) Argc=1;
  else { Argc=argc; Argv=argv; flag_opt=1;
	 optind=save_optind; ansicpp_argc=argc; ansicpp_argv=argv; }
}

if (RodsWidth<0) { // option for TR and width<0
  iWidth2=-long(RodsWidth*GRSizeX*PaperX); iWidth=2*iWidth2; }

if ( DoAutoscaleX || DoAutoscaleY || DoAutoscaleZ )
  AutoScale(DoAutoscaleX,DoAutoscaleY,DoAutoscaleZ);

#if 0
// some verification for ticks --- taken away, 1.3.97 I've added a way default behaviour
if (!XticksOn)
    { XticksFrom=Xfrom; XticksTo=Xto; XticksStep=Xto-Xfrom; }
  else if (XticksTo>Xto) XticksTo=Xto;
if (!YticksOn)
    { YticksFrom=Yfrom; YticksTo=Yto; YticksStep=Yto-Yfrom; }
  else if (YticksTo>Yto) YticksTo=Yto;
#endif

// now test and filled default params if they were not set:

if (InpFile==0)
  { cerr <<"Input data file not specified\n"; exit(1); }

/* NO: now redirection to cerr
if (OutFile==0)
  { OutFile=new char [7]; strcpy(OutFile,"out->ps"); }
*/

if (optind!=argc)
  { cerr <<"Parameter "<<ansicpp_argv[optind]<<" is wrong\n"; exit(1); }

if (!PaperNameCalled)
  { strcpy(charTmp,"a4l"); PaperName2Values(charTmp); }

ShiftX = PaperX*(0.5*(1.-GRSizeX)+GX1+ShiftX0); ShiftY = PaperY*(0.5*(1.-GRSizeY)+GY1+ShiftY0);
ThisEveryNth=1;
/******** end of reading the command line opts ****/

ScaleX=GRSizeX*PaperX/(Xto-Xfrom); ScaleY=GRSizeY*PaperY/(Yto-Yfrom);

/*
Technical note:
  - left bottom corner of the graph is
      [ShiftX,ShiftY],
  - right upper corner of the graph is
      [ShiftX+GRSizeX*PaperX,ShiftY+GRSizeY*PaperY]
*/

// approximate the postscript bounding box
tmp = 0.16;
if (!VerticalYTicksLabels) tmp += 0.03;
BoundBoxXA = ShiftX-tmp*GRSizeX*PaperX;
BoundBoxYA=ShiftY-0.13*GRSizeY*PaperY;
BoundBoxXB=ShiftX+1.1*GRSizeX*PaperX;
BoundBoxYB=ShiftY+1.1*GRSizeY*PaperY;


if (Zto<Zfrom || (isZlinlog && (Zfrom<=0)))
  { cerr <<"Bad boundaries of Z axis.\n"; exit(1); }
if (isZlinlog)
  { Zfrom=log10(Zfrom); Zto=log10(Zto); }
Zinterval = Zto-Zfrom;

Ax = (double*) malloc(sizeof(double)*ArrayMax);
Ay = (double*) malloc(sizeof(double)*ArrayMax);
Az = (double*) malloc(sizeof(double)*ArrayMax);
Ain = (char*) malloc(sizeof(char)*ArrayMax);
Bx = (double*) malloc(sizeof(double)*ArrayMax);
By = (double*) malloc(sizeof(double)*ArrayMax);
Bz = (double*) malloc(sizeof(double)*ArrayMax);
Bin = (char*) malloc(sizeof(char)*ArrayMax);
if (!Ax||!Ay||!Az||!Ain||!Bx||!By||!Bz||!Bin) NotMemory();
StartColor=0; MaxColor=1; // interval of gray
Scans=0;

ifFF {
  Inp.open(InpFile); if (!Inp)
    { cerr <<"Cannot open file " <<InpFile <<" !\n"; exit(1); }
#ifndef __TURBOC__ // some problems
#if !(defined(__GNUC__) && __GNUC__>=3)
  Inp.setbuf(input_buffer,size_input_buffer);
#endif
#endif
  }

// here it would be possible to mark the heap (to avoid memory leaks)

if (OutFile==0 || !strcmp(OutFile,"-")) {
    // redirect to standard output
    #ifdef __DECCXX
      cerr <<"FATAL: Cannot redirect to stdout under VMS/DECCXX, sorry.\n";
      exit(1);
    #elif (defined(__GNUC__) && __GNUC__>=3)
      out = &cout;
    #else
      // (void*)out=(void*)&cout;
      out = new ofstream(fileno(stdin));
    #endif
    // CanCloseOut=0;
    CanCloseOut=1;
    }
  else {
    if (*OutFile!='|') { // ordinary file open
	#ifdef PIPES
	char *pt = strrchr( OutFile, '.' );
	if ( pt && !strcmp(pt,".gz") ) { //.gz file required on output
	      if (!pipes_avail()) {
		cerr <<"FATAL: Cannot .gz output" <<" --- pipes are not available on your system!\n";
		exit(1); }
	      char s[255] = "gzip -c >";
	      strcat( s, OutFile );
	      FILE *outPipe = popen( s, "wb" );
	      if (outPipe==NULL) {
		cerr <<"FATAL: Cannot open gzip for output\n";
		exit(1); }
#ifndef __TURBOC__ // some problems
	      setbuf(outPipe,output_buffer);
#endif
	      out = new ofstream(fileno(outPipe));
	      CanCloseOut = 1;
	      }
	else {
	#endif
	out = new ofstream;
	#ifdef __unix // avoid problems with (wrongly) missing ios::binary
	  ((ofstream*)out)->open(OutFile,ios::out);
	#else
	  ((ofstream*)out)->open(OutFile,ios::out|ios_bin);
	#endif
	if (!*out)
	  { cerr <<"FATAL: Cannot write file " <<OutFile <<" !\n"; exit(1); }
#ifndef __TURBOC__ // some problems
#if !(defined(__GNUC__) && __GNUC__>=3)
	((ofstream*)out)->setbuf(output_buffer,size_output_buffer);
#endif
#endif
	CanCloseOut = 1;
	#ifdef PIPES //add the ending parenthesis for .gz output
	}
	#endif
	}
      else { // open pipe
	#ifdef PIPES
	if (!pipes_avail()) {
	  cerr <<"FATAL: Cannot pipe output" <<" --- pipes are not available on your system!\n";
	  exit(1); }
	PipeLikeFile = popen(OutFile+1,"wb");
	if (PipeLikeFile==0)
	  { cerr <<"FATAL: Cannot write to pipe " <<OutFile+1 <<" !\n"; exit(1); }
#ifndef __TURBOC__ // some problems
	setbuf(PipeLikeFile,output_buffer);
#endif
	// C to C++ conversion
	// out = new ofstream(PipeLikeFile->FILE_HANDLE);
	out = new ofstream( fileno(PipeLikeFile) );
	CanCloseOut = 1;
	#else
	#if (defined(__GNUC__) && __GNUC__>=3)
	cerr <<"FATAL: Your brand new C++ compiler does not support output to pipe, sorry.\n";
	#else
	cerr <<"FATAL: Your obsolete/toonew system/compiler does not allow output to pipes. Upgrade!\n";
	#endif
	exit(1);
	#endif
	}
    }
cerr <<"Reading file " <<InpFile
     <<" and writing ";
switch (ColorMode) {
  case colorModeGRAY: cerr <<"gray"; break;
  case colorModeRGB: cerr <<"colour (rgb "
			  <<formulaR <<';' <<formulaG <<';' <<formulaB <<')'; break;
  default: cerr <<"==Cannot be here!==";
  }
cerr <<" file " <<OutFile <<endl;
WriteHeader();


if (*Title) {
  *out<<"0 G "; font(6000); *out<<"\t\t%title\n";
  if (ExpandText(6000,Title,TextBuffer,"show",theshift))
    *out<<wrtGraphX(theshift)<<' '<<wrtGraphY(1.05)<<" M ("<<TextBuffer<<") show\n";
  else // not expanded text can be centered:
    *out<<wrtGraphX(0.5)<<' '<<wrtGraphY(1.05)<<" M ("<<TextBuffer<<") Cshow\n";

  }
if (*LabelX) {
  tmp = TicksOut ? -0.05: -0.03;
  *out<<"0 G "; font(4500); *out<<"\t\t%label X\n";
  if (ExpandText(4500,LabelX,TextBuffer,"show",theshift))
    *out<<wrtGraphX(theshift)<<' '<<wrtGraphY(tmp)<<" (m) stringwidth pop neg add"
	<<" M ("<<TextBuffer<<") showTopAlign\n";
  else // not expanded text can be centered:
    *out<<wrtGraphX(0.5)<<' '<<wrtGraphY(tmp)<<" (m) stringwidth pop neg add"
	<<" M ("<<TextBuffer<<") CshowTopAlign\n";
  }
  /* old: only two margin labels:
  font(3000); *out<<endl;
  sprintf(LabelX,"%1.8g",Xfrom);
  *out<<wrtGraphX(0)<<' '<<wrtGraphY(-0.035)<<" M ("<<LabelX<<") show\n";
  sprintf(LabelX,"%1.8g",Xto);
  *out<<wrtGraphX(1)<<' '<<wrtGraphY(-0.035)<<" M ("<<LabelX<<") Rshow\n";
  }
  */
Xticks();
if (*LabelY) {
  tmp = TicksOut ? -0.05: -0.03;
  if (!VerticalYTicksLabels) tmp -= 0.03; // 10.11.1998 shift for horiz y ticks labels
  *out<<"0 G "; font(4500); *out<<"\t\t%label Y\n";
  if (ExpandText(4500,LabelY,TextBuffer,"RotateShow",theshift))
    *out<<wrtGraphX(tmp)<<" (m) stringwidth pop neg add "<<wrtGraphY(theshift)
	<<" M ("<<TextBuffer<<") RotateShow\n";
  else // not expanded text can be centered:
    *out<<wrtGraphX(tmp)<<" (m) stringwidth pop neg add "<<wrtGraphY(0.5)
	<<" M ("<<TextBuffer<<") RotateCshow\n";
  }
  /* old: only two margin labels:
  font(3000); *out<<endl;
  sprintf(LabelY,"%1.8g",Yfrom);
  *out<<wrtGraphX(-0.015)<<' '<<wrtGraphY(0)<<" M ("<<LabelY<<") RotateShow\n";
  sprintf(LabelY,"%1.8g",Yto);
  *out<<wrtGraphX(-0.015)<<' '<<wrtGraphY(1)<<" M ("<<LabelY<<") RotateRshow\n";
  }
  */
Yticks();
MakeGridXY();
if (isFFand readnotes && *charTmp) { // read and write notes
  FromLine -= 2;
  Inp.get(charTmp,sizeof(charTmp)-1); Inp.xligno();
  font(3500); *out<<"\t\t%notes\n";
  *out <<"0.1 G "<<wrtGraphX(NotesX)<<' '
      <<wrtGraphY(NotesY)<<" M ("<<charTmp;
  Inp.get(charTmp,sizeof(charTmp)-1); Inp.xligno();
  ExpandText(3500,charTmp,TextBuffer,"show",theshift);
  *out << TextBuffer << ") show\n";
  // Inp.getline(Note; Note2; cat them); write them;
  }

if (isFFand Inp.eof_skipspaces()) goto TheEnd;
if (*OutFile=='|') out->flush();
ifFF { // skip lines
  for (LL=FromLine-Inp.line; LL-->0; )
    { Inp.xligno(); if (!Inp) goto TheEnd; }
  for (LL=FromScan; --LL>0; ) {  // skip scans
    while (Inp.eoln_skipspaces())
      { Inp.xligno(); if (!Inp) goto TheEnd; }
    while (!Inp.eoln_skipspaces())
      { Inp.xligno(); if (!Inp) goto TheEnd; }
    }
  }

//EstablishClip:
*out <<'\n'<<"gsave "<<1-isnegative<<" G 0 w\t\t\t\t%now establish clipping\n";
iGX1=long( floor(x2p(Xfrom)) ); iGY1=long( floor(y2p(Yfrom)) );
iGX2=long( floor(x2p(Xto)) );   iGY2=long( floor(y2p(Yto)) );
*out << iGX1 <<' '<< iGY1 <<" M ";
*out << iGX2-iGX1 <<' '<< 0 <<" V ";
*out << 0 <<' '<< iGY2-iGY1 <<" V ";
*out << iGX1-iGX2 <<' '<< 0 <<" V closepath clip stroke\n";

// read and draw scans until eof
long W;
word ai,bi,aimax,bimax /*, aiup*/;

out->setf(ios::fixed, ios::floatfield);
if (!Rods) *out << "\n0 w";
  else
    if (RodsWidth>0) *out <<'\n';
      else  *out <<'\n' << iWidth << " w";
*out << "\n%pm3d_map_begin\n";

AgainA:;
do { // read the 1st scan to A arrays
  if (ReadScan(aimax)) goto TheEnd;
} while (aimax<2); // accept scan with 2 pts at least
MaxAimax=aimax;
// copy scan B to scan A
W=long(sizeof(double))*aimax;
pmemcpy(Ax,Bx,W); pmemcpy(Ay,By,W); pmemcpy(Az,Bz,W);
FillBin(aimax); pmemcpy(Ain,Bin,aimax);

while ( /*(isFromWhich==isFromFormula) ||*/
   (isFFand !Inp.eof_skipspaces())
    && !(kbhit() && getch()=='\x1B') // stop it by ESC
   ) {
  if (CanCopyScan) { // scan B to scan A
    aimax=bimax;
    W=long(sizeof(double))*aimax;
    pmemcpy(Ax,Bx,W); pmemcpy(Ay,By,W); pmemcpy(Az,Bz,W); }
  CanCopyScan=1;
  // read the next scan
  if (ReadScan(bimax) && bimax<2) goto TheEnd; // last scan, not enough points
  if (bimax<2) { // ignore scans with 1 point only
    CanCopyScan=0;
    goto AgainA; }
  FillBin(bimax);
  MaxAimax = (MaxAimax>bimax) ? MaxAimax:bimax;
  // aiup=MIN(aimax,bimax)-1;

  if (Rods) goto LabelRods;

  // MAKE NORMAL GRID (USING MY 3D ALGORITHM)
  switch (Flush) {
    case 'R': if (aimax<bimax) { ai=0; bi=bimax-aimax; }
		else { bi=0; ai=aimax-bimax; }
	      break;
    case 'C': if (aimax<bimax) { ai=0; bi=int((bimax-aimax)/2); }
		else { bi=0; ai=int((aimax-bimax)/2); }
	      break;
    default /*'L'*/: ai=0; bi=0; }
  for (; ai<aimax-1 && bi<bimax-1 /*ai<aiup*/; ai++, bi++) {
    // bi=ai;
    if (Ain[ai] || Ain[ai+1] || Bin[bi] || Bin[bi+1] ) {
      DoBlankLine=1;
      // for each point of A scan (the prev. scan)
      if (Speed) SetColor( Get3dColori(ai,bi) );
      LineColor=Get3dColori(ai,bi);
      iGX1=(long)x2p(Ax[ai]);   iGY1=(long)y2p(Ay[ai]);
      iGX2=(long)x2p(Bx[bi]);   iGY2=(long)y2p(By[bi]);
      iGX3=(long)x2p(Bx[bi+1]); iGY3=(long)y2p(By[bi+1]);
      iGX4=(long)x2p(Ax[ai+1]); iGY4=(long)y2p(Ay[ai+1]);
      // typeset the rectangle. x and y coordinates can be rounded
      // since our unit is 0.01 bp
      out->precision(0);

      if (Speed) { // output optimized for Speed
	// new version with relative path:
	*out << iGX1<<' '<<iGY1<<" N ";
#if 0 /* version "x1 y1 V x2 y2 V x3 y3 f" */
	*out << iGX2-iGX1<<' '<<iGY2-iGY1<<" V "
	     << iGX3-iGX2<<' '<<iGY3-iGY2<<" V "
	     << iGX4-iGX3<<' '<<iGY4-iGY3<<" f\n";
#else /* version "x3 y3 x2 y2 x1 y1 h" */
	*out << iGX4-iGX3<<' '<<iGY4-iGY3<<' '
	     << iGX3-iGX2<<' '<<iGY3-iGY2<<' '
	     << iGX2-iGX1<<' '<<iGY2-iGY1<<" h\n";
#endif
	/* // older version with relative path:
	out->precision(0);
	*out << floor(GX1)<<' '<<floor(GY1)<<" N ";
	*out << floor(GX2-GX1)<<' '<<floor(GY2-GY1)<<" V ";
	*out << floor(GX3-GX2)<<' '<<floor(GY3-GY2)<<" V ";
	*out << floor(GX4-GX3)<<' '<<floor(GY4-GY3)<<" f\n";
	*/
	/* // the first version with absolute path:
	out->precision(0);
	*out << floor(GX1)<<' '<<floor(GY1)<<" N ";
	*out << floor(GX2)<<' '<<floor(GY2)<<" L ";
	*out << floor(GX3)<<' '<<floor(GY3)<<" L ";
	*out << floor(GX4)<<' '<<floor(GY4)<<" L F\n";
	// *out << floor(GX1)<<' '<<floor(GY1)<<" L X\n";
	*/
	} // of Speed output

      else { // file Size optimization
      #ifndef RelativeRectangles
      sprintf(Line,"%ld %li N %li %li V %li %li V %li %li f",
		 iGX1,iGY1,iGX2-iGX1,iGY2-iGY1,iGX3-iGX2,iGY3-iGY2,iGX4-iGX3,iGY4-iGY3);
      #else
      if (firstsprintf) {
	iGXfix=iGX1; iGYfix=iGY1;
	*out << iGXfix <<' '<<iGYfix<<" M currentpoint\n";
	firstsprintf=0; }
      sprintf(Line,"n %ld %li R %li %li V %li %li V %li %li f",
		 iGX1-iGXfix,iGY1-iGYfix,iGX2-iGX1,iGY2-iGY1,iGX3-iGX2,iGY3-iGY2,iGX4-iGX3,iGY4-iGY3);
      #endif

      //do the last 2 lines concide?
      LineLen = strlen(Line);
      LineEnd = Line+LineLen; //address of \x0

      //cout <<"\n\nOLD LINE="<<OldLine<<endl;  //for debugging
      //cout <<"    LINE="<<Line<<endl;

      if (!isTail) { // tail does not exist
	if (!*OldLine) { //no prev. line in buffer
			 //no print, save Line to OldLine
	    strcpy(OldLine,Line); OldLineLen = LineLen;
	    OldLineEnd = OldLine+OldLineLen; //address of \x0
	    OldLineColor = LineColor;
	    }
	  else { //compare OldLine and Line -- do their tails coincide?
	    cptr = memcmp_rev_space(OldLineEnd-2,LineEnd-2,MIN(OldLineLen,LineLen));
	    if (LineEnd-cptr<6) {
		//ignore Tail since their common tails are shorter than __ chars
		//print OldLine, save Line to OldLine
		if (*OldLine) {
		  SetColor(OldLineColor); *out <<OldLine <<'\n'; }
		strcpy(OldLine,Line); OldLineLen=LineLen;
		OldLineEnd=OldLine+OldLineLen; //address of \x0
		OldLineColor=LineColor;
		}
	      else {
		//introduce Tail, write down its definition,
		//print abbreviated OldLine and Line
		while (*cptr && *cptr==' ') cptr++;  //skip spaces
		strcpy(Tail,cptr); TailLen=strlen(Tail);
		*out << "/x {"<<Tail<<"} D\n";
		OldLine[OldLineLen-TailLen]=0; //cut OldLine
		SetColor(OldLineColor); *out << OldLine << "x\n";
		Line[LineLen-TailLen]=0; //cut Line
		SetColor(LineColor); *out << Line << "x\n";
		isTail=1;
		}
	    }
	}
	else { // Tail exists
	   // do Tail and tail of Line coincide?
	   if (!strcmp(Tail,LineEnd-TailLen)) {
	       // yes, thus print abbreviated Line
	       Line[LineLen-TailLen]=0;
	       SetColor(LineColor); *out << Line << "x\n";
	       // there are no changes in Tail
	       }
	     else {
	      // no, thus undef Tail definition: save Line in OldLine
	      strcpy(OldLine,Line); OldLineLen = LineLen;
	      OldLineEnd = OldLine+OldLineLen; //address of \x0
	      OldLineColor = LineColor;
	      isTail = 0;
	      }
	  }
      } // of Size optimized output
      if (!*out) {
	cerr<<"Problem writing to output file.\n";
	if (++outproblems>=MaxOutProblems) {
	  cerr <<"Too much output problems (disk full? disk quota exceeded?) Terminating.\n";
	  goto TheEnd; }
	}
      }
    } // for the whole A scan, i.e. for each ai
  // END OF MY 3D ALGORITHM
  goto LabelAfterRods;


  // USE RODS ALGORITHM

  LabelRods:;
//FillBin(aimax);
//pmemcpy(Ain,Bin,aimax);
//PMTRs for (ai=bimax; ai--; )
  //PMTRs   if (Ain[ai]) cout<<"+";
//PMTRs PMTRACE("DOING NEXT SCAN: aimax="<<aimax<<"  bimax="<<bimax);
  // Note: do not forget to update the same thing below TheEnd:
  // if (RodsWidth>0) iWidth2=iWidth/2; //it would be for relative width
  for (ai=0; ai<aimax; ai++) {  // aiup
  bi = ai;
  if (Ain[ai] || (ai+1<aimax && Ain[ai+1]) || (ai>0 && Ain[ai-1])) {
  DoBlankLine = 1;

  SetColor(Get3dColor(Az[ai]));

  if (ai==0) {
      iGX1 = (long)x2p(Ax[0]);   iGY1 = (long)y2p(Ay[0]); //ai point
      iGX2 = (long)x2p(Ax[1]);   iGY2 = (long)y2p(Ay[1]); //next point
      iGX4 = iGX1+(iGX1-iGX2)/2; iGY4 = iGY1+(iGY1-iGY2)/2; //current point of pen
      }
    else {
      iGX1=iGX2; iGY1=iGY2; } //ai point
  if (ai<aimax-1) {
      iGX2 = (long)x2p(Ax[ai+1]); iGY2 = (long)y2p(Ay[ai+1]); //next point
      iGX3 = (iGX2+iGX1)/2; iGY3 = (iGY2+iGY1)/2; } //next line pt=mean(X1,X4)

  if (RodsWidth>0) { // recalculate the width
    if (ai<bimax) {
      iWidth = (int) sqrt( sqr(iGX1-(long)x2p(Bx[ai])) + sqr(iGY1-(long)y2p(By[ai])) );
      if (iWidth!=OldiWidth) {
	*out << iWidth*RodsWidth << " w "; OldiWidth=iWidth; }
      }
    }
  if (ai<aimax-1) {
      *out << iGX3-iGX4 <<' '<< iGY3-iGY4 << ' '
	   << iGX4 <<' '<< iGY4 << " v\n";
      iGX4=iGX3; iGY4=iGY3; // new current point
      }
    else //prolongated line
      *out << 2*(iGX1-iGX4) << ' ' << 2*(iGY1-iGY4) << ' '
	   << iGX4 <<' '<< iGY4 << " v\n";
  } // for each point in the drawn area
  } // for this/each scan
  // END OF RODS ALGORITHM

  LabelAfterRods:;
  if (bimax>0) pmemcpy(Ain,Bin,bimax);
  if (DoBlankLine) *out <<'\n'; // blank line - new scan
  DoBlankLine=0;
  if (*OutFile=='|') out->flush();
  } // for all scans, i.e. until eof or ESC
TheEnd:

if (Rods) {
  // the last, remaining scan is stored in B fields
  aimax=bimax;
  // Below is the copy of LabelRods: above, but with A fields replaced by B fields
  // if (RodsWidth>0) iWidth2=iWidth/2; //it would be for relative width
  for (ai=0; ai<aimax; ai++) {  // aiup
   bi=ai;
  if (Bin[ai] || (ai+1<aimax && Bin[ai+1]) || (ai>0 && Bin[ai-1])) {
  DoBlankLine=1;

  SetColor(Get3dColor(Bz[ai]));

  if (ai==0) {
      iGX1=(long)x2p(Bx[0]); iGY1=(long)y2p(By[0]); //ai point
      iGX2=(long)x2p(Bx[1]); iGY2=(long)y2p(By[1]); //next point
      iGX4=iGX1+(iGX1-iGX2)/2; iGY4=iGY1+(iGY1-iGY2)/2; //current point of pen
      }
    else {
      iGX1=iGX2; iGY1=iGY2; } //ai point
  if (ai<aimax-1) {
      iGX2=(long)x2p(Bx[ai+1]); iGY2=(long)y2p(By[ai+1]); //next point
      iGX3=(iGX2+iGX1)/2; iGY3=(iGY2+iGY1)/2; } //next line pt=mean(X1,X4)

  if (RodsWidth>0) { // recalculate the width
    if (ai<aimax) {
      iWidth=(int)sqrt(sqr(iGX1-(long)x2p(Ax[ai]))+sqr(iGY1-(long)y2p(Ay[ai])));
      if (iWidth!=OldiWidth) {
	*out << iWidth*RodsWidth << " w "; OldiWidth=iWidth; }
      }
    }
  if (ai<aimax-1) {
      *out << iGX3-iGX4 <<' '<< iGY3-iGY4 << ' '
	   << iGX4 <<' '<< iGY4 << " v\n";
      iGX4=iGX3; iGY4=iGY3; // new current point
      }
    else //prolongated line
      *out << 2*(iGX1-iGX4) << ' ' << 2*(iGY1-iGY4) << ' '
	   << iGX4 <<' '<< iGY4 << " v\n";
  }
  }
  } // end of Rods and last scan and for each point

TerminateOut();
// it is possible now to release the heap
// free [] Ax, Ay, Az, Ain;
// free [] Bx, By, Bz, Bin;
return;
} // end of Main Part: GrafFrom3D


void Help ( int exit_code )
{
cerr <<"(mikulik@physics.muni.cz, http://www.sci.muni.cz/~mikulik/)\n";
cerr <<"Program to draw a 3d data file as a gray/colour scale map.\n";

cerr <<"\nThis program is free software. It is distributed in the hope that it\n";
cerr <<"will be useful,  but WITHOUT ANY WARRANTY;  without even the implied\n";
cerr <<"warranty of  MERCHANTABILITY  or  FITNESS FOR A PARTICULAR PURPOSE.\n";
cerr <<"See the GNU General Public License for more details.\n\n\n";

#ifdef __VMS
cerr << "  Usage on VAX/VMS systems:  pm3d:==\"$disk:[dir]pm3d.exe\"\n";
cerr << "                             pm3d [pm3d_input_file] [options]\n";
#else
cerr << "  Usage:   pm3d  pm3d_input_file [options]\n";
#endif
cerr <<"\nThe input file is  pm3d.p3d  file compatible (see the distribution).\n";
cerr <<"There are 2 possible formats of *.p3d files, please use the new one now:\n";
cerr <<"  1. the old one (the only format till March 1997), and\n";
cerr <<"  2. the new format, in addition supported by Jpm3d---the Java GUI for pm3d.\n";
cerr <<"\n  Options overwriting setup in pm3d_input_file:\n";
cerr <<"-S               ... optimize size of the output file\n";
cerr <<"-f size          ... relative font size (default 1.0)\n";
cerr <<"-a combination   ... autoscale, combinations allowed are x,y,z,xy,yz,xz,xyz\n";
cerr <<"-i inpfile       ... input 3d data file (at least 3 columns, each scan is\n";
cerr <<"                     separated by a blank line; see the Demo directory)\n";
cerr <<"-o outfile       ... output postscript file\n";
cerr <<"-g x y           ... relative graph sizes -s x y  ... additional shifts\n";
cerr <<"-l               ... log/linear Z axis    -n      ... negative/positive image\n";
cerr <<"-b               ... no boundary          -B      ... black uncovered regions\n";
cerr <<"-r fromline fromscan everynth             -R w  truncation_rod_algorithm width\n";
cerr <<"-X formula4x        -Y formula4y          -Z formula4z\n";
cerr <<"-x xfrom xto        -y yfrom yto          -z zfrom zto\n";
cerr <<"-tx from step to ... ticks on x axis      -ty from step to ... ticks on y axis\n";
cerr <<"-P paper         ... paper can be a4 or a4l or dos or dosl\n";
cerr <<"-FL, -FR, -FC    ... flush scans Left (default), Right or Center\n";
cerr <<"-G formula4gray  ... formula for the gray scale transformation\n";
cerr <<"-L labelx labely ... labels for axes (in quotes if each longer than 1 word)\n";
cerr <<"-T title         ... title of the graph (in quotes if longer than 1 word)\n";
cerr <<"-N [N | Y posx posy] write the first two lines of the file as notes at this pos\n";
cerr <<"-K show_n_gray_levels x y  ... relative position wrt graph\n";
cerr <<"-o- ... redirect output to stdout     -w ... wait to Ctrl/C (kill) before close\n";
cerr <<"-o\"|prog\"  ... pipe postscript into prog (i.e. \"-o|gs -\"); Unix and OS/2 only\n";
cerr <<"-c colormode ... define colour mode (r for rgb, default g for gray)\n";
cerr <<"-C colorparam... parameter for the colour R mode (e.g. 2;-1;4), default "<<formulaR<<';'<<formulaG<<';'<<formulaB<<'\n';
cerr <<"\t\t Other nice:  3;11;6  23;28;3  3;28;23  30,31,32  33;13;10\n";
cerr <<"\nThe distribution of this program (C++ source code) is accesible from my Web\n";
cerr <<"page, see above.\n";
exit(exit_code);
}



// ******************** MAIN *************************

int main ( int argc, char *argv[] )
{
clock_t start_clock = clock();
cerr <<"\nProgram  pm3d  (version " <<TheDate <<") - c " <<Petr <<" - Brno\n";

if (argc==1) Help(0);

ansicpp_argc=argc; ansicpp_argv=argv;
GrafFrom3D ( argc, argv );

double tmp = clock() - start_clock;
tmp /= CLOCKS_PER_SEC;
cerr <<"Elapsed time: " <<setprecision(3) <<tmp <<" s = " <<tmp/60 <<" min\n";
cerr <<"Program  pm3d - end.\n";
return 0;
}


// eof pm3d.cpp
