// PMParser.Cpp: formula parser
// c Petr Mikulk - 1994

#include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include "pmparser.h"

char ParserError;    // global error indicator of parser
char *ParserIndex;   // address of char where an error occured
			    // if ParserError != 0

char ParserUsexyz, ParserUseDollars, ParserDollardouble;
parserfloat *ParserDollars;

char   *( ParserConstName[] ) =
       {"pi","e","rad2deg","deg2rad","rad2sec","sec2rad",""};
double ParserConst     []    =
       { 3.14159265358979323846,2.71828182845904523536,
	 180./3.14159265358979323846,3.14159265358979323846/180,
	 6480000./3.14159265358979323846,3.14159265358979323846/6480000.
       };


double parser_sinc ( double x ) {
 if (!x) return 1; else return sin(x)/x; }
double parser_sgn ( double x )
  { if (!x) return 0; else return x>0 ? 1 : -1; }
double parser_sqr ( double x ) { return x*x; }
double parser_log ( double x ) { return x>0 ? log(x):-1e308; }
double parser_log10 ( double x ) { return x>0 ? log10(x):-1e308; }
double parser_step ( double x ) { return x>=0 ? 1 : 0; }
double parser_int ( double x ) { return int(x); }
double parser_floor ( double x ) { return floor(x); }
double parser_ceil ( double x ) { return ceil(x); }
double parser_round ( double x ) { return floor(x+0.5); }
double parser_random ( double x ) { return rand()/RAND_MAX*x; }
double parser_srand ( double x ) { srand(int(x)); return 0; }

char *( ParserFuncName[] ) =
       {"sin","cos","tan","log","log10","exp","abs","fabs",
	"sqrt","sqr","asin","acos","atan","sinh","cosh","tanh",
	"sinc","sgn","step","int","floor","ceil","round",
	"random","srand",""};
double (*ParserFunc[])(double) =
      { sin,cos,tan,parser_log,parser_log10,exp,fabs,fabs,
	sqrt,parser_sqr,asin,acos,atan,sinh,cosh,tanh,
	parser_sinc,parser_sgn,parser_step,parser_int,
	parser_ceil,parser_ceil,parser_round,parser_random,parser_srand };

double parser_eq ( double x, double y ) { return x==y; }
double parser_mod ( double x, double y ) { return x-y*floor(x/y); }

char *( ParserFunc2Name[] ) =
       {"pow","eq","mod",""};
double (*ParserFunc2[])(double,double) =
      { pow,parser_eq,parser_mod };

double ifno0 ( double a, double x, double y ) { return a ? x : y; }
double ifpos ( double a, double x, double y ) { return a>0 ? x : y; }
double ifpos0 ( double a, double x, double y ) { return a>=0 ? x : y; }
double ifneg ( double a, double x, double y ) { return a<0 ? x : y; }
double ifneg0 ( double a, double x, double y ) { return a<=0 ? x : y; }

char *( ParserFunc3Name[] ) =
       {"ifno0","ifpos","ifpos0","ifneg","ifneg0",""};
double (*ParserFunc3[])(double,double,double) =
      { ifno0,ifpos,ifpos0,ifneg,ifneg0 };


void nospaces ( char *S ) {
char *i,*j; for (i=j=S; *i; i++) if (!isspace(*i)) *j++=*i; *j=0;
}


// remark: DoFormulaParser and ParserEvalExpr are called recursively
parserfloat DoFormulaParser ( char **Index );


parserfloat ParserEvalExpr ( char **Index, char UnaryEnabled ) {
int Unary=1;
if (UnaryEnabled)
  switch (**Index) { case '+': (*Index)++; break;
  case '-': (*Index)++; Unary=-1; }
char *Index0=*Index; char Old; parserfloat X;
// read number
if (isdigit(**Index) || **Index=='.') {
  X=strtod(*Index,&Index0);
  *Index=Index0;
  return Unary>0 ? X : -X;
  }
int i;
// read $n parameter
if (ParserUseDollars && **Index=='$') {
  (*Index)++;
  if (!isdigit(**Index) || (isdigit(*(*Index+1))) && isdigit(*(*Index+2)))
    { ParserError=peDollar; return 0; }
  i=*((*Index)++) - '0';
  if (isdigit(**Index)) i=10*i+*((*Index)++)-'0';
  if (ParserDollardouble) X=ParserDollars[i];
    else X=((float*)ParserDollars)[i];
  return Unary>0 ? X : -X;
  }
int var=1; // nb of variables(params) in parenth.
if (**Index!='(') {
  // read name of constant or function
  if (!isalnum(**Index))
    { ParserError=peIdentifier; return 0; }
  while ( isalnum(*(++*Index)) );
  // is it x,y or z? They are equiv. to $1,$2,$3
  if ( ParserUsexyz && *Index-Index0==1 &&
       *Index0>='x' && *Index0<='z' )
    { i=*Index0-'w';
      if (ParserDollardouble) X=ParserDollars[i];
        else X=((float*)ParserDollars)[i];
      return Unary>0 ? X : -X;
    }
  Old=**Index; **Index=0;
  if (Old!='(') { // it is constant
    for (i=0; strcmp(Index0,ParserConstName[i]); i++)
      if (!ParserConstName[i][0]) // no such constant
        { ParserError=peIdentifier; return 0; }
    X=ParserConst[i];
    **Index=Old;
    return Unary>0 ? X : -X;
    }
  // it is function
  for (i=0; strcmp(Index0,ParserFuncName[i]); i++)
    if (!ParserFuncName[i][0]) // no such fun of 1 param.
    { var=2;
      for (i=0; strcmp(Index0,ParserFunc2Name[i]); i++)
      if (!ParserFunc2Name[i][0]) // no such fun of 2 param.
      { var=3;
        for (i=0; strcmp(Index0,ParserFunc3Name[i]); i++)
        if (!ParserFunc3Name[i][0]) // no such fun of 3 param.
        { ParserError=peIdentifier; return 0; }
        break;
      }
      break;
    }
  **Index=Old;
  }
// evaluate the expression(s) in the parenthesis
double X1,X2;
for (int v=var; v; v--) {
char *Index1, *Index2; Index1=Index2=++(*Index);   // position of '('+1
for (int k=1; k; Index2++) // find left parentheses or comma
  switch (*Index2) {
    case ')': k--; break;
    case '(': k++; break;
    case ',': if (k==1 && v>1) k--; break; // not last param
    case 0  : ParserError=peParentheses; return 0;
    }
char Old2=*(--Index2); *Index2=0;
X=DoFormulaParser( &Index1 );
if (v==2) X1=X; else if (v==3) X2=X;
*Index2=Old2; // it was ) or ,
if (ParserError) return 0;
*Index=Index2+1;
if (v>1) --(*Index);
}

if (*Index0!='(') // it was function and not only parenthesis
  switch (var) {
    case 1: X=ParserFunc[i](X); break;
    case 2: X=ParserFunc2[i](X1,X); break;
    case 3: X=ParserFunc3[i](X2,X1,X); break;
    }
return Unary>0 ? X : -X;
}


parserfloat DoFormulaParser ( char **Index ) {
// routine which is called by FormulaParser or recursively via EvalExpr
// It performs  Expr1 {+= | -=} {Expr2 {{*= | /=} Expr3} } till end or error
//                     ...=Oper1         ...=Oper2
parserfloat Expr1=0, Expr2, Expr3;
char   Oper1='+', Oper2=' ';
char   UnaryEnable=1;
while (Oper2) {
  Expr2=ParserEvalExpr( Index, UnaryEnable );
  if (ParserError) return 0;
  UnaryEnable=0;
  Oper2=**Index;
  if (!Oper2) goto StopIt;
  if (!strchr("+-*/",Oper2))
    { ParserError=peOperator; return Expr1; }
  (*Index)++;
  while (strchr("*/",Oper2)) {
    // loop of multiplication and division
    Expr3=ParserEvalExpr(Index,0);
    if (ParserError) return Expr1;
    switch (Oper2) {
      case '*': Expr2 *= Expr3; break;
      case '/': if (Expr3) Expr2 /= Expr3;
                  else Expr2 = Expr2==0 ? 1 : Expr2>0 ? 1e308:-1e308;
      }
    Oper2=**Index;
    if (!Oper2) goto StopIt;
    if (!strchr("+-*/\0",Oper2))
      { ParserError=peOperator; return Expr1; }
    (*Index)++;
    }
  // no more multiplication of division between Expr2 and Expr3,
  // hence add or substract Expr1 and Expr2
  StopIt:;
  switch (Oper1) {
    case '+': Expr1 += Expr2; break;
    case '-': Expr1 -= Expr2; break;
    }
  Oper1 = Oper2;
  }
return Expr1;
} // DoFormulaParser


void FormulaParserPreProcess ( char *S ) {
if (!S) return;
nospaces(S);
#if (defined(__TURBOC__) && !defined(__STDC__)) || defined(__WATCOMC__)
strlwr(S);
#else
int i=0; while ( S[i]=tolower(S[i]) ) i++;
#endif
}

parserfloat FormulaParser ( char *S ) {
if (!S) return 0;
ParserError=0; ParserUsexyz=ParserUseDollars=0;
ParserIndex=S;
return DoFormulaParser ( &ParserIndex );
}


parserfloat FormulaParserDollars ( char *S, parserfloat *Dollars, char isdouble ) {
if (!S) return 0;
ParserError=0; ParserUsexyz=ParserUseDollars=1;
ParserDollars=Dollars; ParserDollardouble=isdouble;
ParserIndex=S;
return DoFormulaParser ( &ParserIndex );
}


int FormulaParserGetMaxParam ( char *S ) {
if (!S) return 0;
char *Index=S; char i, Max=0;
for ( ; *Index; Index++ ) { // find max of $n and $nm
  if ((*Index=='$' && !isdigit((*Index+1))) &&
       ( !isalnum(*Index+2) ||
         ((isdigit(*(Index+2) && !isalnum(*Index+3)))) ))
    { i=(*(++Index)) - '0';
      if (isdigit(*(Index+1))) i=10*i+*(++Index)-'0';
      if (i>Max) Max=i;
    }
  }
// find max of x, y or z
char isfirst=1;
for ( Index=S ; *Index; Index++ ) {
  if (isfirst || !isalnum(*(Index-1)))
    if (*Index>='x' && *Index<='z' && !isalnum(*(Index+1)))
      { i=*Index-'w'; if (i>Max) Max=i; }
  isfirst=0;
  }
return Max;
}


int Formula2Column ( char *S ) {
if (!S || !*S) return 0;
if (!*(S+1) && *S>='x' && *S<='z') return *S-'w';
if (*S!='$') return -1;
if (!isdigit(*++S)) return -1;
if (!*(S+1)) return *S-'0';
if (!isdigit(*(S+1)) || *(S+2)) return -1;
return 10*(*S-'0') + *(S+1)-'0';

}


// eof pmparser.cpp
