// PMParser.Cpp: mathematical formula parser
// (c) Petr Mikulik (mikulik@sci.muni.cz) - 1994
// Freeware version. It is distributed without any warranty.

#ifndef	 PMPARSER_H
#define	 PMPARSER_H


/* Parsing works recursively. It works as interpreter.
   Maybe the code is not clear, but it works.
   This program is public domain if my name remains above.
   It should be distributed together with the demo program.
   New versions (if any) could be found on
   ftp.muni.cz:pub/muni.cz/physics (accessible also via gopher.muni.cz)
*/


typedef double parserfloat;
//typedef float  parserfloat;
  // which  precision is used


// error_in_formula management
extern char ParserError;    // global error indicator of parser
// #define peOK=0
#define peIdentifier   1
#define peOperator     2
#define peParentheses  3
#define peDollar       4
extern char *ParserIndex;   // address of char where an error occured
                     // if ParserError != 0

void FormulaParserPreProcess ( char *S );
// preprocessing. It converts string S to lower case and it removes spaces
// It should be called before the following two routines

parserfloat FormulaParser ( char *S );
// you give string in, like "7.4*sin(pi/3*5.5)+(1-3)*3",
// and it returns the result

parserfloat FormulaParserDollars
  ( char *S, parserfloat *Dollars, char isdouble );
// like the previous one. Moreover, you can address the elements
// of the Dollars array by  $n  names ($0==Dollars[0], $1==Dollars[1] etc.)
// Moreover, keywords x,y,z are equivalents to $1, $2, $3
// Example: "sin( ($1+$2)*deg2rad )*z + 1"

// For the list of the known constants and functions, look to
// the pmparser.cpp file. This should be known:
/* Constants:
     "pi","e","rad2deg","deg2rad","rad2sec","sec2rad"
   Functions of one parameter:
     "sin","cos","tan","log","log10","exp","abs","fabs",
     "sqrt","sqr","asin","acos","atan","sinh","cosh","tanh",
     "sinc","sgn","step","int","floor","ceil","round",
     "random","srand"
     Note: sinc(x)=sin(x)/x
	   step(x)= x>=0 ? 1:0 (Heaviside function)
	   srand(seed) sets the seed of the random generator, returns 0
	   random(x) returns a random number between 0 and x
   Functions of two parameters:
     "pow"  pow(x,y)=x**y
     "eq"   eq(x,y) =x==y
     "mod"  mod(x,y) = x modulo y
   Functions of three parameters:
     ifno0(expr,x,y):   expr!=0 ? x : y
     ifpos(expr,x,y):   expr>0 ? x : y
     ifpos0(expr,x,y):  expr>=0 ? x : y
     ifneg(expr,x,y):   expr<0 ? x : y
     ifneg0(expr,x,y):  expr<=0 ? x : y

   Note: division 0/0 gives 1, division x/0 gives sgn(x)*1e308,
     log,log10 of x<=0 gives -1e308
*/

int FormulaParserGetMaxParam ( char *S );
// returns max. nb. of parameter ($n) in the formula S
// E.g. FormulaParserGetMaxParam( "sin($4)+$1" ) returns 4

int Formula2Column ( char *S );
// it returns 0 if S==null
// it returns 1 if S==x or $1, 2 if S==y or $2,
//            3 if S==z or $3, 4 if S==$5 etc.
// it returns -1 if the formula is not a simple column expression

#endif

// eof pmparser.h
